// -*- C++ -*-
/***************************************************************************
                          packets.cpp  -  description
                             -------------------
    begin                : Mon Jun 11 2001
    copyright            : (C) 2001 by Daniel Hermann
    email                : Daniel.Hermann@stud.uni-karlsruhe.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>

#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "log.h"
#include "mem.h"

#include "packets.h"

/**************************************************************************
Swap bytes on an integer considered as 16 bits
**************************************************************************/
static unsigned int swab_int16(unsigned int val)
{
  return ((val & 0xFF) << 8) | ((val & 0xFF00) >> 8);
}

/**************************************************************************
Swap bytes on an integer considered as 32 bits
**************************************************************************/
static unsigned int swab_int32(unsigned int val)
{
  return ((val & 0xFF) << 24) | ((val & 0xFF00) << 8)
    | ((val & 0xFF0000) >> 8) | ((val & 0xFF000000) >> 24);
}

/**************************************************************************
Swap bytes on an pointed to integer considered as 16 bits
**************************************************************************/
#ifdef UNUSED
static void swab_pint16(int *ptr)
{
  (*ptr) = swab_int16(*ptr);
}
#endif

/**************************************************************************
Swap bytes on an pointed to integer considered as 32 bits
**************************************************************************/
static void swab_pint32(int *ptr)
{
  (*ptr) = swab_int32(*ptr);
}

/**************************************************************************
...
**************************************************************************/
void *get_packet_from_connection(struct connection *pc, int *ptype)
{
  int len, type;

  if(pc->buffer.ndata<4)
    return NULL;           /* length and type not read */

  get_int16(pc->buffer.data, &len);
  get_int8(pc->buffer.data+2, &type);

  if(pc->first_packet) {
    /* the first packet better be short: */
    freelog(LOG_DEBUG, "first packet type %d len %d", type, len);
    if(len > 1024) {
      freelog(LOG_NORMAL, "connection from %s detected as old byte order",
	      pc->addr);
      pc->byte_swap = 1;
    } else {
      pc->byte_swap = 0;
    }
    pc->first_packet = 0;
  }

  if(pc->byte_swap) {
    len = swab_int16(len);
    /* so the packet gets processed (removed etc) properly: */
    put_int16(pc->buffer.data, len);
  }

  if(len > pc->buffer.ndata)
    return NULL;           /* not all data has been read */

  freelog(LOG_DEBUG, "packet type %d len %d", type, len);

  *ptype=type;

  switch(type) {

  case PACKET_REQUEST_JOIN_GAME:
    return receive_packet_req_join_game(pc);

  case PACKET_JOIN_GAME_REPLY:
    return receive_packet_join_game_reply(pc);

  case PACKET_SERVER_SHUTDOWN:
    return receive_packet_generic_message(pc);

  case PACKET_GAME_STATE:
    return receive_packet_generic_integer(pc);


  case PACKET_YOUR_TURN:
  case PACKET_TURN_DONE:
  case PACKET_PAY_REQUEST:
  case PACKET_DECLARE_BANKRUPT:
  case PACKET_CONFIRM_ANIM_FINISHED:
  case PACKET_ANIM_FINISHED:
    return receive_packet_generic_integer(pc);

  case PACKET_DICE_ROLL_REQUEST:
    return receive_packet_generic_integer(pc);
  case PACKET_DICE_ROLLED:
    return receive_packet_dice_rolled(pc);

  case PACKET_CARD_DRAWN:
    return receive_packet_generic_values(pc);
    
  case PACKET_PROPERTY_BUY:
  case PACKET_PROPERTY_MORTGAGE:
  case PACKET_PROPERTY_UNMORTGAGE:
    return receive_packet_generic_integer(pc);

  case PACKET_HOUSES_BUY:
  case PACKET_HOUSES_SELL:
    return receive_packet_house_request(pc);

  case PACKET_JAIL_FREE_REQUEST:
    return receive_packet_generic_integer(pc);

  case PACKET_TRADE:
    return receive_packet_trade(pc);

  case PACKET_FIELD_INFO:
    return receive_packet_field_info(pc);

  case PACKET_MONOP_INFO:
    return receive_packet_monop_info(pc);


  case PACKET_PLAYER_INIT_INFO:
    return receive_packet_player_init_info(pc);

  case PACKET_PLAYER_INFO:
    return receive_packet_player_info(pc);


  case PACKET_GAME_INIT_INFO:
    return receive_packet_game_init_info(pc);

  case PACKET_GAME_INFO:
    return receive_packet_game_info(pc);


  case PACKET_CHAT_MSG:
  case PACKET_PAGE_MSG:
/*    case PACKET_SHOW_MESSAGE: */
    return receive_packet_generic_message(pc);
    
  default:
    freelog(LOG_NORMAL, "unknown packet type received");
    remove_packet_from_buffer(&pc->buffer);
    return NULL;
  };
}

/**************************************************************************
...
**************************************************************************/
unsigned char *get_int8(unsigned char *buffer, int *val)
{
  if(val) {
    *val=(*buffer);
  }
  return buffer+1;
}

/**************************************************************************
...
**************************************************************************/
unsigned char *put_int8(unsigned char *buffer, int val)
{
  *buffer++=val&0xff;
  return buffer;
}

/**************************************************************************
...
**************************************************************************/
unsigned char *get_int16(unsigned char *buffer, int *val)
{
  if(val) {
    unsigned short x;
    memcpy(&x,buffer,2);
    *val=ntohs(x);
  }
  return buffer+2;
}


/**************************************************************************
...
**************************************************************************/
unsigned char *put_int16(unsigned char *buffer, int val)
{
  unsigned short x = htons(val);
  memcpy(buffer,&x,2);
  return buffer+2;
}


/**************************************************************************
...
**************************************************************************/
unsigned char *get_int32(unsigned char *buffer, int *val)
{
  if(val) {
    unsigned long x;
    memcpy(&x,buffer,4);
    *val=ntohl(x);
  }
  return buffer+4;
}


/**************************************************************************
...
**************************************************************************/
unsigned char *put_int32(unsigned char *buffer, int val)
{
  unsigned long x = htonl(val);
  memcpy(buffer,&x,4);
  return buffer+4;
}

/**************************************************************************
...
**************************************************************************/
unsigned char *put_string(unsigned char *buffer, char *mystring)
{
  int len = strlen(mystring) + 1;
  memcpy(buffer, mystring, len);
  return buffer+len;
}


/**************************************************************************
...
**************************************************************************/
unsigned char *get_string(unsigned char *buffer, char *mystring)
{
  unsigned char *c;
  int len;

  /* avoid using strlen (or strcpy) on an (unsigned char*)  --dwp */
  for(c=buffer; *c; c++) ;
  len = c-buffer+1;
  if(mystring) {
    memcpy(mystring, buffer, len);
  }
  return buffer+len;
}



/**************************************************************************
...
**************************************************************************/
int send_packet_generic_message(struct connection *pc, int type,
				struct packet_generic_message *packet)
{
  unsigned char buffer[MAX_PACKET_SIZE], *cptr;
  cptr=put_int8(buffer+2, type);
  cptr=put_int8(cptr, packet->x);
  cptr=put_int8(cptr, packet->y);
  cptr=put_int32(cptr, packet->event);

  cptr=put_string(cptr, packet->message);
  put_int16(buffer, cptr-buffer);

  return send_connection_data(pc, buffer, cptr-buffer);
}


/**************************************************************************
...
**************************************************************************/
struct packet_generic_message *
receive_packet_generic_message(struct connection *pc)
{
  unsigned char *cptr;
  struct packet_generic_message *packet=
    (struct packet_generic_message *) 
    cap_malloc(sizeof(struct packet_generic_message));

  cptr=get_int16(pc->buffer.data, NULL);
  cptr=get_int8(cptr, NULL);
  cptr=get_int8(cptr, &packet->x);
  cptr=get_int8(cptr, &packet->y);
  cptr=get_int32(cptr, &packet->event);

  cptr=get_string(cptr, packet->message);
  
  remove_packet_from_buffer(&pc->buffer);

  return packet;
}


/**************************************************************************
...
**************************************************************************/
int send_packet_generic_integer(struct connection *pc, int type,
				struct packet_generic_integer *packet)
{
  unsigned char buffer[MAX_PACKET_SIZE], *cptr;
  cptr=put_int8(buffer+2, type);
  cptr=put_int32(cptr, packet->value);
  put_int16(buffer, cptr-buffer);
  return send_connection_data(pc, buffer, cptr-buffer);
}

/**************************************************************************
...
**************************************************************************/
struct packet_generic_integer *
receive_packet_generic_integer(struct connection *pc)
{
  unsigned char *cptr;
  struct packet_generic_integer *packet=
    (struct packet_generic_integer *) 
    cap_malloc(sizeof(struct packet_generic_integer));

  cptr=get_int16(pc->buffer.data, NULL);
  cptr=get_int8(cptr, NULL);
  cptr=get_int32(cptr, &packet->value);

  remove_packet_from_buffer(&pc->buffer);

  return packet;
}

/**************************************************************************
...
**************************************************************************/
int send_packet_generic_values(struct connection *pc, int type,
			       struct packet_generic_values *req)
{
  unsigned char buffer[MAX_PACKET_SIZE], *cptr;
  
  cptr=put_int8(buffer+2, type);
  cptr=put_int16(cptr, req->id);
  cptr=put_int32(cptr, req->value1);
  cptr=put_int32(cptr, req->value2);

  put_int16(buffer, cptr-buffer);

  return send_connection_data(pc, buffer, cptr-buffer);
}

/**************************************************************************
...
**************************************************************************/
struct packet_generic_values *
receive_packet_generic_values(struct connection *pc)
{
  unsigned char *cptr;
  int length;
  struct packet_generic_values *packet=
    (struct packet_generic_values *) 
    cap_malloc(sizeof(struct packet_generic_values));

  cptr=get_int16(pc->buffer.data, &length);
  cptr=get_int8(cptr, NULL);
  cptr=get_int16(cptr, &packet->id);
  if(pc->buffer.data+length-cptr >= 4)
    cptr=get_int32(cptr, &packet->value1);
  else packet->value1=0;
  if(pc->buffer.data+length-cptr >= 4)
    cptr=get_int32(cptr, &packet->value2);
  else packet->value2=0;

  remove_packet_from_buffer(&pc->buffer);

  return packet;
}




/**************************************************************************
...
**************************************************************************/
int send_packet_req_join_game(struct connection *pc, struct 
			      packet_req_join_game *request)
{
  unsigned char buffer[MAX_PACKET_SIZE], *cptr;
  cptr=put_int8(buffer+2, PACKET_REQUEST_JOIN_GAME);
  cptr=put_string(cptr, request->name);
  cptr=put_int32(cptr, request->major_version);
  cptr=put_int32(cptr, request->minor_version);
  cptr=put_int32(cptr, request->patch_version);
  put_int16(buffer, cptr-buffer);

  return send_connection_data(pc, buffer, cptr-buffer);
}

/**************************************************************************
...
**************************************************************************/
struct packet_req_join_game *
receive_packet_req_join_game(struct connection *pc)
{
  unsigned char *cptr;
  struct packet_req_join_game *packet=
    (struct packet_req_join_game *)
    cap_malloc(sizeof(struct packet_req_join_game));

  cptr=get_int16(pc->buffer.data, NULL);
  cptr=get_int8(cptr, NULL);
  cptr=get_string(cptr, packet->name);
  cptr=get_int32(cptr, &packet->major_version);
  cptr=get_int32(cptr, &packet->minor_version);
  cptr=get_int32(cptr, &packet->patch_version);

  if(pc->byte_swap) {
    swab_pint32(&packet->major_version);
    swab_pint32(&packet->minor_version);
    swab_pint32(&packet->patch_version);
  }

  remove_packet_from_buffer(&pc->buffer);

  return packet;
}


/**************************************************************************
...
**************************************************************************/
int send_packet_join_game_reply(struct connection *pc, struct 
			        packet_join_game_reply *reply)
{
  unsigned char buffer[MAX_PACKET_SIZE], *cptr;
  cptr=put_int8(buffer+2, PACKET_JOIN_GAME_REPLY);
  cptr=put_int32(cptr, reply->you_can_join);
  /* if other end is byte swapped, you_can_join should be 0,
     which is 0 even if bytes are swapped! --dwp */
  cptr=put_string(cptr, reply->message);

  /* so that old clients will understand the reply: */
  if(pc->byte_swap) {
    put_int16(buffer, swab_int16(cptr-buffer));
  } else {
    put_int16(buffer, cptr-buffer);
  }

  return send_connection_data(pc, buffer, cptr-buffer);
}

/**************************************************************************
...
**************************************************************************/
struct packet_join_game_reply *
receive_packet_join_game_reply(struct connection *pc)
{
  unsigned char *cptr;
  struct packet_join_game_reply *packet=
    (struct packet_join_game_reply *)
    cap_malloc(sizeof(struct packet_join_game_reply));

  cptr=get_int16(pc->buffer.data, NULL);
  cptr=get_int8(cptr, NULL);
  
  cptr=get_int32(cptr, &packet->you_can_join);
  cptr=get_string(cptr, packet->message);

  remove_packet_from_buffer(&pc->buffer);

  return packet;
}


/*************************************************************************
...
**************************************************************************/
int send_packet_dice_rolled(struct connection *pc, 
			    struct packet_dice_rolled *pack)
{
  unsigned char buffer[MAX_PACKET_SIZE], *cptr;

  cptr=put_int8(buffer+2, PACKET_DICE_ROLLED);

  cptr=put_int8(cptr, pack->playerid);
  cptr=put_int8(cptr, pack->dice1);
  cptr=put_int8(cptr, pack->dice2);
/*    cptr=put_int8(cptr, pack->newloc); */
  
  put_int16(buffer, cptr-buffer);

  return send_connection_data(pc, buffer, cptr-buffer);
}


/*************************************************************************
...
**************************************************************************/
struct packet_dice_rolled *
receive_packet_dice_rolled(struct connection *pc)
{
  unsigned char *cptr;
  int length;
  struct packet_dice_rolled *pack=
    (struct packet_dice_rolled *)
     cap_malloc(sizeof(struct packet_dice_rolled));

  cptr=get_int16(pc->buffer.data, &length);
  cptr=get_int8(cptr, NULL);

  cptr=get_int8(cptr, &pack->playerid);
  cptr=get_int8(cptr, &pack->dice1);
  cptr=get_int8(cptr, &pack->dice2);
/*    cptr=get_int8(cptr, &pack->newloc); */
  
  remove_packet_from_buffer(&pc->buffer);
  return pack;
}

/*************************************************************************
...
**************************************************************************/
int send_packet_house_request(struct connection *pc, int type,
			      struct packet_house_request *pack)
{
  unsigned char buffer[MAX_PACKET_SIZE], *cptr;
  int i;

  cptr=put_int8(buffer+2, type);

  cptr=put_int8(cptr, pack->monopid);

  for (i=0; i<MAX_STREETS_IN_MONOP; ++i)
    cptr=put_int32(cptr, pack->nhouses[i]);
  
  put_int16(buffer, cptr-buffer);

  return send_connection_data(pc, buffer, cptr-buffer);
}


/*************************************************************************
...
**************************************************************************/
struct packet_house_request *
receive_packet_house_request(struct connection *pc)
{
  unsigned char *cptr;
  int length, i;
  struct packet_house_request *pack=
    (struct packet_house_request *)
     cap_malloc(sizeof(struct packet_house_request));

  cptr=get_int16(pc->buffer.data, &length);
  cptr=get_int8(cptr, NULL);

  cptr=get_int8(cptr, &pack->monopid);

  for (i=0; i<MAX_STREETS_IN_MONOP; ++i)
    cptr=get_int32(cptr, &pack->nhouses[i]);
  
  remove_packet_from_buffer(&pc->buffer);
  return pack;
}


/*************************************************************************
...
**************************************************************************/
int send_packet_player_init_info(struct connection *pc, 
				 struct packet_player_init_info *pinfo)
{
  unsigned char buffer[MAX_PACKET_SIZE], *cptr;

  cptr=put_int8(buffer+2, PACKET_PLAYER_INIT_INFO);

  
  cptr=put_int8(cptr, pinfo->playerid);
  cptr=put_string(cptr, pinfo->name);
  cptr=put_string(cptr, pinfo->addr);
  
  put_int16(buffer, cptr-buffer);

  return send_connection_data(pc, buffer, cptr-buffer);
}


/*************************************************************************
...
**************************************************************************/
struct packet_player_init_info *
receive_packet_player_init_info(struct connection *pc)
{
  unsigned char *cptr;
  int length;
  struct packet_player_init_info *pinfo=
    (struct packet_player_init_info *)
    cap_malloc(sizeof(struct packet_player_init_info));

  cptr=get_int16(pc->buffer.data, &length);
  cptr=get_int8(cptr, NULL);

  cptr=get_int8(cptr, &pinfo->playerid);
  cptr=get_string(cptr, pinfo->name);
  cptr=get_string(cptr, pinfo->addr);
  
  remove_packet_from_buffer(&pc->buffer);
  return pinfo;
}


/*************************************************************************
...
**************************************************************************/
int send_packet_player_info(struct connection *pc, 
			    struct packet_player_info *pinfo)
{
  int i;
  unsigned char buffer[MAX_PACKET_SIZE], *cptr;

  cptr=put_int8(buffer+2, PACKET_PLAYER_INFO);

  cptr=put_int8(cptr, pinfo->playerid);
  cptr=put_int32(cptr, pinfo->event);
  cptr=put_int32(cptr, pinfo->statusflags);
  cptr=put_int32(cptr, pinfo->actionflags);

  cptr=put_int8(cptr, pinfo->location);
  cptr=put_int8(cptr, pinfo->nturns_in_jail);

  cptr=put_int32(cptr, pinfo->budget);
  cptr=put_int32(cptr, pinfo->transaction);
  cptr=put_int8(cptr, pinfo->partner);

  for (i=0; i<BOARD_N_FIELDS; ++i) {
    cptr=put_int8(cptr, pinfo->holdings[i]);
  }

  cptr=put_int8(cptr, pinfo->ncards);

  cptr=put_int32(cptr, pinfo->bank_debts);
  for (i=0; i<GAME_MAX_PLAYERS; ++i)
    cptr=put_int32(cptr, pinfo->debts[i]);
  
  put_int16(buffer, cptr-buffer);

  return send_connection_data(pc, buffer, cptr-buffer);
}


/*************************************************************************
...
**************************************************************************/
struct packet_player_info *
receive_packet_player_info(struct connection *pc)
{
  int i;
  unsigned char *cptr;
  int length;
  struct packet_player_info *pinfo=
    (struct packet_player_info *)
    cap_malloc(sizeof(struct packet_player_info));

  cptr=get_int16(pc->buffer.data, &length);
  cptr=get_int8(cptr, NULL);

  cptr=get_int8(cptr, &pinfo->playerid);
  cptr=get_int32(cptr, &pinfo->event);
  cptr=get_int32(cptr, &pinfo->statusflags);
  cptr=get_int32(cptr, &pinfo->actionflags);

  cptr=get_int8(cptr, &pinfo->location);
  cptr=get_int8(cptr, &pinfo->nturns_in_jail);

  cptr=get_int32(cptr, &pinfo->budget);
  cptr=get_int32(cptr, &pinfo->transaction);
  cptr=get_int8(cptr, &pinfo->partner);

  for (i=0; i<BOARD_N_FIELDS; ++i) {
    cptr=get_int8(cptr, &pinfo->holdings[i]);
  }

  cptr=get_int8(cptr, &pinfo->ncards);

  cptr=get_int32(cptr, &pinfo->bank_debts);
  for (i=0; i<GAME_MAX_PLAYERS; ++i)
    cptr=get_int32(cptr, &pinfo->debts[i]);
  
  remove_packet_from_buffer(&pc->buffer);
  return pinfo;
}

/*************************************************************************
...
**************************************************************************/
int send_packet_field_info(struct connection *pc, 
			   struct packet_field_info *pack)
{
  unsigned char buffer[MAX_PACKET_SIZE], *cptr;

  cptr=put_int8(buffer+2, PACKET_FIELD_INFO);

  cptr=put_int8(cptr, pack->location);
  cptr=put_int32(cptr, pack->event);
  cptr=put_int32(cptr, pack->statusflags);
  cptr=put_int32(cptr, pack->actionflags);
  cptr=put_int8(cptr, pack->owner);
  cptr=put_int8(cptr, pack->nhouses);

  put_int16(buffer, cptr-buffer);

  return send_connection_data(pc, buffer, cptr-buffer);
}


/*************************************************************************
...
**************************************************************************/
struct packet_field_info *
receive_packet_field_info(struct connection *pc)
{
  unsigned char *cptr;
  int length;
  struct packet_field_info *pack=
    (struct packet_field_info *)
    cap_malloc(sizeof(struct packet_field_info));

  cptr=get_int16(pc->buffer.data, &length);
  cptr=get_int8(cptr, NULL);

  cptr=get_int8(cptr, &pack->location);
  cptr=get_int32(cptr, &pack->event);
  cptr=get_int32(cptr, &pack->statusflags);
  cptr=get_int32(cptr, &pack->actionflags);
  cptr=get_int8(cptr, &pack->owner);
  cptr=get_int8(cptr, &pack->nhouses);

  remove_packet_from_buffer(&pc->buffer);
  return pack;
}

/*************************************************************************
...
**************************************************************************/
int send_packet_monop_info(struct connection *pc, 
			   struct packet_monop_info *pack)
{
  unsigned char buffer[MAX_PACKET_SIZE], *cptr;

  cptr=put_int8(buffer+2, PACKET_MONOP_INFO);

  cptr=put_int8(cptr, pack->monopid);
  cptr=put_int32(cptr, pack->event);
  cptr=put_int8(cptr, pack->owner);

  put_int16(buffer, cptr-buffer);

  return send_connection_data(pc, buffer, cptr-buffer);
}


/*************************************************************************
...
**************************************************************************/
struct packet_monop_info *
receive_packet_monop_info(struct connection *pc)
{
  unsigned char *cptr;
  int length;
  struct packet_monop_info *pack=
    (struct packet_monop_info *)
    cap_malloc(sizeof(struct packet_monop_info));

  cptr=get_int16(pc->buffer.data, &length);
  cptr=get_int8(cptr, NULL);

  cptr=get_int8(cptr, &pack->monopid);
  cptr=get_int32(cptr, &pack->event);
  cptr=get_int8(cptr, &pack->owner);

  remove_packet_from_buffer(&pc->buffer);
  return pack;
}


/*************************************************************************
...
**************************************************************************/
int send_packet_game_init_info(struct connection *pc, 
			       struct packet_game_init_info *ginfo)
{
  int i,j;
  unsigned char buffer[MAX_PACKET_SIZE], *cptr;

  cptr=put_int8(buffer+2, PACKET_GAME_INIT_INFO);

  
  for (i=0; i<BOARD_N_FIELDS; ++i) {
    cptr=put_int8(cptr, ginfo->board.field[i].location);
    cptr=put_int8(cptr, ginfo->board.field[i].type);
    cptr=put_int32(cptr, ginfo->board.field[i].price);
    cptr=put_int32(cptr, ginfo->board.field[i].house_price);
    for (j=0; j<STREET_N_RENT_CLASSES; ++j) {
      cptr=put_int32(cptr, ginfo->board.field[i].rent[j]);
    }
  }    
  for (i=0; i<BOARD_N_MONOPS; ++i) {
    cptr=put_int8(cptr, ginfo->board.monop[i].monopid);
    cptr=put_int8(cptr, ginfo->board.monop[i].nstreets);
    for (j=0; j<MAX_STREETS_IN_MONOP; ++j) {
      cptr=put_int8(cptr, ginfo->board.monop[i].street[j]);
    }
  }
    
  cptr=put_int8(cptr, ginfo->nplayers);
  cptr=put_int8(cptr, ginfo->your_plrid);
  
  for (i=0; i<ginfo->nplayers; ++i) {
    cptr=put_int8(cptr, ginfo->player[i].playerid);
    cptr=put_string(cptr, ginfo->player[i].name);
    cptr=put_string(cptr, ginfo->player[i].addr);
  }
  
  put_int16(buffer, cptr-buffer);

  return send_connection_data(pc, buffer, cptr-buffer);
}


/*************************************************************************
...
**************************************************************************/
struct packet_game_init_info *
receive_packet_game_init_info(struct connection *pc)
{
  int i,j;
  unsigned char *cptr;
  int length;
  struct packet_game_init_info *ginfo=
    (struct packet_game_init_info *)
    cap_malloc(sizeof(struct packet_game_init_info));

  cptr=get_int16(pc->buffer.data, &length);
  cptr=get_int8(cptr, NULL);

  for (i=0; i<BOARD_N_FIELDS; ++i) {
    cptr=get_int8(cptr, &ginfo->board.field[i].location);
    cptr=get_int8(cptr, &ginfo->board.field[i].type);
    cptr=get_int32(cptr, &ginfo->board.field[i].price);
    cptr=get_int32(cptr, &ginfo->board.field[i].house_price);
    for (j=0; j<STREET_N_RENT_CLASSES; ++j) {
      cptr=get_int32(cptr, &ginfo->board.field[i].rent[j]);
    }
  }    
  for (i=0; i<BOARD_N_MONOPS; ++i) {
    cptr=get_int8(cptr, &ginfo->board.monop[i].monopid);
    cptr=get_int8(cptr, &ginfo->board.monop[i].nstreets);
    for (j=0; j<MAX_STREETS_IN_MONOP; ++j) {
      cptr=get_int8(cptr, &ginfo->board.monop[i].street[j]);
    }
  }
    
  cptr=get_int8(cptr, &ginfo->nplayers);
  cptr=get_int8(cptr, &ginfo->your_plrid);

  for (i=0; i<ginfo->nplayers; ++i) {
    cptr=get_int8(cptr, &ginfo->player[i].playerid);
    cptr=get_string(cptr, ginfo->player[i].name);
    cptr=get_string(cptr, ginfo->player[i].addr);
  }
  
  remove_packet_from_buffer(&pc->buffer);
  return ginfo;
}


/*************************************************************************
...
**************************************************************************/
int send_packet_game_info(struct connection *pc, 
			  struct packet_game_info *ginfo)
{
  int i,j;
  unsigned char buffer[MAX_PACKET_SIZE], *cptr;

  cptr=put_int8(buffer+2, PACKET_GAME_INFO);

  cptr=put_int32(cptr, ginfo->event);
  
  for (i=0; i<BOARD_N_FIELDS; ++i) {
    cptr=put_int8(cptr, ginfo->board.field[i].location);
    cptr=put_int32(cptr, ginfo->board.field[i].event);
    cptr=put_int32(cptr, ginfo->board.field[i].statusflags);
    cptr=put_int32(cptr, ginfo->board.field[i].actionflags);
    cptr=put_int8(cptr, ginfo->board.field[i].owner);
    cptr=put_int8(cptr, ginfo->board.field[i].nhouses);
  }
  for (i=0; i<BOARD_N_MONOPS; ++i) {
    cptr=put_int8(cptr, ginfo->board.monop[i].monopid);
    cptr=put_int32(cptr, ginfo->board.monop[i].event);
    cptr=put_int8(cptr, ginfo->board.monop[i].owner);
  }

  cptr=put_int8(cptr, ginfo->nplayers);
  
  for (i=0; i<ginfo->nplayers; ++i) {
    cptr=put_int8(cptr, ginfo->player[i].playerid);
    cptr=put_int32(cptr, ginfo->player[i].event);
    cptr=put_int32(cptr, ginfo->player[i].statusflags);
    cptr=put_int32(cptr, ginfo->player[i].actionflags);

    cptr=put_int8(cptr, ginfo->player[i].location);
    cptr=put_int8(cptr, ginfo->player[i].nturns_in_jail);

    cptr=put_int32(cptr, ginfo->player[i].budget);
    cptr=put_int32(cptr, ginfo->player[i].transaction);
    cptr=put_int8(cptr, ginfo->player[i].partner);
    
    for (j=0; j<BOARD_N_FIELDS; ++j) {
      cptr=put_int8(cptr, ginfo->player[i].holdings[j]);
    }    
    cptr=put_int8(cptr, ginfo->player[i].ncards);

    cptr=put_int32(cptr, ginfo->player[i].bank_debts);
    for (j=0; j<GAME_MAX_PLAYERS; ++j)
      cptr=put_int32(cptr, ginfo->player[i].debts[j]);
  
  }

  put_int16(buffer, cptr-buffer);

  return send_connection_data(pc, buffer, cptr-buffer);
}


/*************************************************************************
...
**************************************************************************/
struct packet_game_info *
receive_packet_game_info(struct connection *pc)
{
  int i,j;
  unsigned char *cptr;
  int length;
  struct packet_game_info *ginfo=
    (struct packet_game_info *)
    cap_malloc(sizeof(struct packet_game_info));

  cptr=get_int16(pc->buffer.data, &length);
  cptr=get_int8(cptr, NULL);

  cptr=get_int32(cptr, &ginfo->event);

  for (i=0; i<BOARD_N_FIELDS; ++i) {
    cptr=get_int8(cptr, &ginfo->board.field[i].location);
    cptr=get_int32(cptr, &ginfo->board.field[i].event);
    cptr=get_int32(cptr, &ginfo->board.field[i].statusflags);
    cptr=get_int32(cptr, &ginfo->board.field[i].actionflags);
    cptr=get_int8(cptr, &ginfo->board.field[i].owner);
    cptr=get_int8(cptr, &ginfo->board.field[i].nhouses);
  }
  for (i=0; i<BOARD_N_MONOPS; ++i) {
    cptr=get_int8(cptr, &ginfo->board.monop[i].monopid);
    cptr=get_int32(cptr, &ginfo->board.monop[i].event);
    cptr=get_int8(cptr, &ginfo->board.monop[i].owner);
  }

  cptr=get_int8(cptr, &ginfo->nplayers);
  
  for (i=0; i<ginfo->nplayers; ++i) {
    cptr=get_int8(cptr, &ginfo->player[i].playerid);
    cptr=get_int32(cptr, &ginfo->player[i].event);
    cptr=get_int32(cptr, &ginfo->player[i].statusflags);
    cptr=get_int32(cptr, &ginfo->player[i].actionflags);

    cptr=get_int8(cptr, &ginfo->player[i].location);
    cptr=get_int8(cptr, &ginfo->player[i].nturns_in_jail);

    cptr=get_int32(cptr, &ginfo->player[i].budget);
    cptr=get_int32(cptr, &ginfo->player[i].transaction);
    cptr=get_int8(cptr, &ginfo->player[i].partner);

    for (j=0; j<BOARD_N_FIELDS; ++j) {
      cptr=get_int8(cptr, &ginfo->player[i].holdings[j]);
    }    

    cptr=get_int8(cptr, &ginfo->player[i].ncards);

    cptr=get_int32(cptr, &ginfo->player[i].bank_debts);
    for (j=0; j<GAME_MAX_PLAYERS; ++j)
      cptr=get_int32(cptr, &ginfo->player[i].debts[j]);

  }

  remove_packet_from_buffer(&pc->buffer);
  return ginfo;
}








/*************************************************************************
...
**************************************************************************/
int send_packet_trade(struct connection *pc, struct packet_trade *pack)
{
  int i;
  unsigned char buffer[MAX_PACKET_SIZE], *cptr;

  cptr=put_int8(buffer+2, PACKET_TRADE);

  cptr=put_int32(cptr, pack->tradeid);

  cptr=put_int8(cptr, pack->state);
  cptr=put_int32(cptr, pack->event);

  cptr=put_int8(cptr, pack->srcid);
  cptr=put_int8(cptr, pack->destid);

  cptr=put_int8(cptr, pack->offer_from);

  cptr=put_int32(cptr, pack->money);
  
  for (i=0; i<BOARD_N_FIELDS; ++i) {
    cptr=put_int32(cptr, pack->prop[i]);
  }

  cptr=put_int8(cptr, pack->cards);
  
  put_int16(buffer, cptr-buffer);

  return send_connection_data(pc, buffer, cptr-buffer);
}


/*************************************************************************
...
**************************************************************************/
struct packet_trade *receive_packet_trade(struct connection *pc)
{
  int i;
  unsigned char *cptr;
  int length;
  struct packet_trade *pack= 
    (struct packet_trade *) cap_malloc(sizeof(struct packet_trade));

  cptr=get_int16(pc->buffer.data, &length);
  cptr=get_int8(cptr, NULL);

  cptr=get_int32(cptr, &pack->tradeid);

  cptr=get_int8(cptr, &pack->state);
  cptr=get_int32(cptr, &pack->event);

  cptr=get_int8(cptr, &pack->srcid);
  cptr=get_int8(cptr, &pack->destid);

  cptr=get_int8(cptr, &pack->offer_from);

  cptr=get_int32(cptr, &pack->money);
  
  for (i=0; i<BOARD_N_FIELDS; ++i) {
    cptr=get_int32(cptr, &pack->prop[i]);
  }

  cptr=get_int8(cptr, &pack->cards);


  remove_packet_from_buffer(&pc->buffer);
  return pack;
}




/**************************************************************************
remove the packet from the buffer
**************************************************************************/
void remove_packet_from_buffer(struct socket_packet_buffer *buffer)
{
  int len;
  get_int16(buffer->data, &len);
  memcpy(buffer->data, buffer->data+len, buffer->ndata-len);
  buffer->ndata-=len;
}

/********************************************************************
...
********************************************************************/
void connection_do_buffer(struct connection *pc)
{
  if(pc)
    pc->send_buffer.do_buffer_sends++;
}

/********************************************************************
...
********************************************************************/
void connection_do_unbuffer(struct connection *pc)
{
  if(pc) {
    pc->send_buffer.do_buffer_sends--;
    if(pc->send_buffer.do_buffer_sends==0)
      flush_connection_send_buffer(pc);
  }
}

/********************************************************************
  flush'em
********************************************************************/
void flush_connection_send_buffer(struct connection *pc)
{
  if(pc) {
    if(pc->send_buffer.ndata) {
      if(write(pc->sock, pc->send_buffer.data, pc->send_buffer.ndata)!=
	 pc->send_buffer.ndata) {
	freelog(LOG_NORMAL, "failed writing data to socket");
      }
      pc->send_buffer.ndata=0;
    }
  }
}


/********************************************************************
  write data to socket
********************************************************************/
int send_connection_data(struct connection *pc, unsigned char *data, int len)
{
  if(pc) {
    if(pc->send_buffer.do_buffer_sends) {
      if(10*MAX_PACKET_SIZE-pc->send_buffer.ndata >= len) { /* room for more?*/
	memcpy(pc->send_buffer.data+pc->send_buffer.ndata, data, len);
	pc->send_buffer.ndata+=len;
      }
      else {
	flush_connection_send_buffer(pc);
	memcpy(pc->send_buffer.data, data, len);
	pc->send_buffer.ndata=len;
      }
    }
    else {
      if(write(pc->sock, data, len)!=len) {
	freelog(LOG_NORMAL, "failed writing data to socket");
	return -1;
      }
      
    }

  }
  return 0;
}



/********************************************************************
read data from socket, and check if a packet is ready

returns: -1  an error occured - you should close the socket
          >0 : #bytes read 
********************************************************************/
int read_socket_data(int sock, struct socket_packet_buffer *buffer)
{
  int didget;

  if((didget=read(sock, buffer->data+buffer->ndata, 
		  MAX_PACKET_SIZE-buffer->ndata))<=0)
    return -1;
 
  buffer->ndata+=didget;
  
  freelog(LOG_DEBUG, "didget:%d", didget); 

  return didget;
}


