// -*- C++ -*-
/***************************************************************************
                          game.h  -  description
                             -------------------
    begin                : Fre Jun 22 17:19:03 CEST 2001
    copyright            : (C) 2001 by Daniel Hermann
    email                : Daniel.Hermann@stud.uni-karlsruhe.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef GAME_H
#define GAME_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "global.h"
#include "shared.h"

#include "player.h"
#include "trade.h"

class Board;
class Field;
class Monop;

class Game {
  
public:

  /*
   * server constructor
   */
  Game();

  /*
   * client constructor
   */
  Game(struct packet_game_init_info *);

  /*
   * destructor
   */
  ~Game();
  
  Board * const board() const { return _board; }
  Player * const player(int id) const 
    { assert(!nobody(id)); return _player[id]; }

  const int nplayers() const { return _nplayers; }

  const int startbudget() const { return _startbudget; }
  const int income_on_GO() const { return _income_on_GO; }

  bool is_on_turn(Player *plr)
    { return (_curr_plr == plr->id()); }
  bool nobody(int plrid) const
    { return (plrid <0 || plrid >= _nplayers); }

  const int die1() const
    { return _last_dice[0]; }
  const int die2() const
    { return _last_dice[1]; }

  const int last_dice_total() const
    { return _last_dice[0]+_last_dice[1]; }

  const bool anim_in_progress() const;

  void set_anim_finished(const int plrid, const bool finished = true);

  int add_player(Player *player);

  int remove_player(const int plrno);
  int remove_all_players();

  Player *find_player_by_name(char *name);
  
  int run();


  /*
   * network input handler (server side)
   */
  void handle_turn_done(Player *);
  void handle_dice_roll_request(Player *, struct packet_generic_integer *);
  
  void handle_property_buy(Player *, struct packet_generic_integer *);
  void handle_property_mortgage(Player *, struct packet_generic_integer *);
  void handle_property_unmortgage(Player *, struct packet_generic_integer *);

  void handle_houses_buy(Player *, struct packet_house_request *);
  void handle_houses_sell(Player *, struct packet_house_request *);

  void handle_jail_free_request(Player *, struct packet_generic_integer *);
  void handle_pay_request(int player_no);
  void handle_declare_bankrupt(Player *);

  void handle_trade_request(Player *, struct packet_trade *);

  void handle_anim_finished(Player *);

  /*
   * server send methods (public)
   */
  void send_your_turn(Player *dest, int plr_id);
  void send_dice_rolled(Player *dest, int dicevalue[]);
  void send_card_drawn(Player *dest, Player *src, 
		       int cardtype, int card_no);
  void send_trade(Player *dest, Trade *, int event, int offer_from);

  void send_game_init_info(Player *dest);
  void send_game_info(Player *dest);

  void send_player_init_info(Player *dest, Player *obj);
  void send_player_info(Player *dest, Player *obj, int event);

  void send_field_info(Player *dest, Field *, int event);
  void send_monop_info(Player *dest, Monop *, int event);

  void send_all_info(Player *dest);

  // ptype is packet type (begin, end, confirm_anim)
  void send_anim(Player *dest, int ptype, int anim_id = 0);

  int complement_event(int event) const;

  /*
   * game flow
   */
  void next_turn();
  void pre_dice();	// not implemented yet
  void post_dice();	// not implemented yet
  void end_turn();

  void roll_dice(int dicevalue[] = 0, const int dice_sum = 0);

  void no_2nd_turn()
    { _ndoubles = 0; }

  bool plr_have_trade(Player *p1, Player *p2)
    { return (_trade[p1->id()][p2->id()] || _trade[p2->id()][p1->id()]); }

  Trade *get_trade(Player *p1, Player *p2)
    {
      Trade *t = _trade[p1->id()][p2->id()];
      if (!t) t = _trade[p2->id()][p1->id()];
      else assert(_trade[p2->id()][p1->id()] == NULL);
      return t;
    }

  void set_trade(Player *p1, Player *p2, Trade *t)
    {
      assert(!plr_have_trade(p1, p2));
      _trade[p1->id()][p2->id()] = t;
      p1->set_trade_with(p2, t);
      p2->set_trade_with(p1, t);
    }

  void clear_trade(Player *p1, Player *p2)
    {
      if (_trade[p1->id()][p2->id()]) delete _trade[p1->id()][p2->id()];
      if (_trade[p2->id()][p1->id()]) delete _trade[p2->id()][p1->id()];
      _trade[p1->id()][p2->id()] = NULL;
      _trade[p2->id()][p1->id()] = NULL;
      p1->clear_trade_with(p2);
      p2->clear_trade_with(p1);
    }

protected:

  Board *_board;

  /*
   * TODO: Spieler nach einer eindeutigen ID numerieren, nicht nach Feldindex
   */
  Player *_player[GAME_MAX_PLAYERS];
  Trade *_trade[GAME_MAX_PLAYERS][GAME_MAX_PLAYERS];

  int _nplayers;
  int _startbudget;
  int _income_on_GO;

  int _randseed; 

  int _curr_plr;  // current player on turn
  int _last_dice[2];
  bool _rolled_dice;	// true, if current player has rolled the dice yet
  int _ndoubles;	// number of times current player rolled doubles 
			// this turn
  bool _turn_done;

  bool _plr_anim_finished[GAME_MAX_PLAYERS];

protected:

//    Player *& player_by_id(int id) { return _player[id]; }

  void buffer_all_connections();
  void unbuffer_all_connections();

  /*
   * generate infos (server side)
   */
  void init_info_for(Player *player, struct packet_game_init_info *) const;
  void info_for(Player *player, struct packet_game_info *) const;


  int check_trade_packet_consistency(Player *, struct packet_trade *);

  void wait_for_anim_finished();

};

void handle_game_init_info(struct packet_game_init_info *packet);


#endif  /* GAME_H */



