// -*- C++ -*-
/***************************************************************************
                          monop.cpp  -  description
                             -------------------
    begin                : Fre Jun 22 17:19:03 CEST 2001
    copyright            : (C) 2001 by Daniel Hermann
    email                : Daniel.Hermann@stud.uni-karlsruhe.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <assert.h>

#include "events.h"
#include "log.h"

#include "plrhand.h"

#include "game.h"
#include "street.h"
#include "player.h"

#include "monop.h"

extern Game *game;

Monop::Monop(int id, Field *field[], const int street_idx[])
  : _id(id),
    _nstreets(0),
    _owner(NULL)
{
  int i, idx;

  // choose 2 or 3 streets in field[], indexed by street_idx

  for (i=0; i<MAX_STREETS_IN_MONOP; ++i) {
    idx = street_idx[i];
    if (idx == -1) break;
    assert(idx >= 0 || idx <= BOARD_N_FIELDS);
    _street[i] = dynamic_cast<Street*>(field[idx]);
    assert(_street[i]);
    _nstreets++;
  }

  // set this Monop in the streets
  for (i=0; i<_nstreets; ++i)
    _street[i]->set_monop(this);
}

bool Monop::check()
{
  int i;
  Player *new_owner = _street[0]->owner();
  
  freelog(LOG_NORMAL, "check monopoly %d", id());

  for (i=1; i<_nstreets; ++i) 
    if (_street[i]->owner() != new_owner) {
      new_owner = NULL;
      break;
    }

  if (_owner != new_owner) {	// Monop has changed
    int xevent = ( (new_owner==NULL) ? E_MONOP_DESTROYED 
		   : (_owner==NULL)  ? E_NEW_MONOP
		   : E_MONOP_CHANGED );

    _owner = new_owner;

    // update street->_is_monop
    for (i=0; i<_nstreets; ++i) {
      _street[i]->set_is_monop((_owner!=NULL));
      game->send_field_info(NULL, _street[i], xevent);
    }

    game->send_monop_info(NULL, this, xevent);
    freelog(LOG_NORMAL, "is monopoly", id());
    return true;
  }
  freelog(LOG_NORMAL, "no monopoly");
  return false;
}


int Monop::buy_houses(Player *plr, int addhouses[])
{
  int newhouses[_nstreets];
  int i, j;
  int nhouses_total = 0, cost;

  if (!is_yours(plr)) {
    notify_player(plr, "Game: you have no monopoly there to build houses");
    freelog(LOG_NORMAL, "error: has no monopoly to build houses");
    return 1;
  } 

  for (i=0; i<_nstreets; ++i) {
    newhouses[i] = _street[i]->nhouses() + addhouses[i];

    if (newhouses[i] > STREET_MAX_HOUSES + 1) {  // maximum of 4 houses + hotel
	notify_player(plr, ("Game: you can only have a maximum of %d houses "
			    "plus 1 hotel on each street"), STREET_MAX_HOUSES);
	freelog(LOG_NORMAL, ("error: can only have a maximum of %d houses "
			     "plus 1 hotel on each street"), 
		STREET_MAX_HOUSES);
	return 1;
    }
    if (newhouses[i] < 0) {  // could happen, if we sell too much houses
      notify_player(plr, ("Game: you can't sell more houses than you have."));
      freelog(LOG_NORMAL, ("error: can't sell more houses than they have"));
      return 1;
    }

    for (j=0; j<i; ++j)
      if ( abs(newhouses[i]-newhouses[j]) > 1 ) {
	notify_player(plr, "Game: you must build houses more balanced");
	freelog(LOG_NORMAL, "error: must build houses more balanced");
	return 1;
      }
  }
  
  // has player enough money ?
  for (i=0; i<_nstreets; ++i)
    nhouses_total += addhouses[i];
  // we assume here that house cost is equal on all streets in each monopoly
  cost = nhouses_total * _street[0]->house_price();

  if ( !plr->has_enough_money(cost) ) {
    notify_player(plr, ("Game: you haven't got enough money to buy houses:\n"
			"\t %d * %d houses  =  %d"), 
		  _street[0]->house_price(), nhouses_total, cost);
    freelog(LOG_NORMAL, "error: has not enough money to buy houses");
    return 1;
  }
  if (cost>0 && plr->debts_sum() > 0) {
    // first settle debts before spending money
    notify_player(plr, "Game: You are in debts. Please settle them before "
		  "spending money.");
    freelog(LOG_NORMAL, "error: is in debts");
  }


  // can't buy houses on mortgaged streets 
  for (i=0; i<_nstreets; ++i) {
    if (_street[i]->is_mortgaged()) {	
      notify_player(plr, "Game: you can't buy houses in monopoly %d with "
		    "mortgaged street %d",
		    id(), _street[i]->location());
      freelog(LOG_NORMAL, "error: can't buy houses in monopoly %d with "
	      "mortgaged street %d",
	      id(), _street[i]->location());
      return 1;
    }
  }

  // note: tests in street::buy_houses are unsafe because if there is a fault
  // in the last street processed, there will already be houses built on
  // first street, so we have tested everything here

  // OK, buy houses for each street separately 
  // but begin with selling of houses (to get money first)

  for (i=0; i<_nstreets; ++i) {
    if (addhouses[i]<0) {	// addhouses[i] could be zero -> ignore it
      _street[i]->sell_houses(plr, -addhouses[i]);
    }
  }
  // ... then buy houses
  for (i=0; i<_nstreets; ++i) {
    if (addhouses[i]>0) {	// addhouses[i] could be zero -> ignore it
      _street[i]->buy_houses(plr, addhouses[i]);
    }
  }
  
  // send additional field infos (after house deals are finished)
  // TODO: these field infos should be the only field infos
  // currently there are still field infos in buy/sell_houses, because
  // client is used to this (every field info coupled with player info)
  for (i=0; i<_nstreets; ++i) {
    // send update for all fields (not only these where houses have changed
      game->send_field_info(TOALL, _street[i], E_NOEVENT);
  }

  // TODO: E_DEAL_HOUSES
  game->send_monop_info(NULL, this, E_BUY_HOUSES);

  return 0;
}

int Monop::sell_houses(Player *plr, int subhouses[])
{
  int newhouses[_nstreets];
  int i, j;

  if (!is_yours(plr)) {
    notify_player(plr, "Game: you have no monopoly there to sell houses");
    freelog(LOG_NORMAL, "error: has no monopoly to build houses");
    return 1;
  } 

  for (i=0; i<_nstreets; ++i) {
    newhouses[i] = _street[i]->nhouses() - subhouses[i];

    if (newhouses[i] < 0) {  // can't sell more houses than exist
      notify_player(plr, "Game: you can only sell as many houses as exist");
      freelog(LOG_NORMAL, "error: can only sell as many houses as exist");
      return 1;
    }      

    for (j=0; j<i; ++j)
      if ( abs(newhouses[i]-newhouses[j]) > 1 ) {	// unbalanced
	notify_player(plr, "Game: you must sell houses more balanced");
	freelog(LOG_NORMAL, ("error: must sell houses more balanced"));
	return 1;
      }
  }
  
  // no tests in Street::sell_houses (see buy_houses) !

  // OK, sell houses for each street separately

  for (i=0; i<_nstreets; ++i) {
    if (subhouses[i]>0) {	// subhouses[i] could be zero -> ignore it
      _street[i]->sell_houses(plr, subhouses[i]);
    }
  }
  
  // send additional field infos (after house deals are finished)
  // TODO: these field infos should be the only field infos
  // currently there are still field infos in buy/sell_houses, because
  // client is used to this (every field info coupled with player info)
  for (i=0; i<_nstreets; ++i) {
    // send update for all fields (not only these where houses have changed
      game->send_field_info(TOALL, _street[i], E_NOEVENT);
  }

  game->send_monop_info(NULL, this, E_SELL_HOUSES);

  return 0;
}

int Monop::sell_one_house(Player *plr)
{
  // find street with largest number of houses 
  int max_houses = 0;
  int street_with_max_houses = 0;
  for (int i=0; i<_nstreets; ++i) {
    if (_street[i]->nhouses() > max_houses) {
      max_houses = _street[i]->nhouses();
      street_with_max_houses = i;
    }
  }
  // sell house on that street (if any house)
  if (max_houses > 0) {
    _street[street_with_max_houses]->sell_houses(plr, 1);
  }

  // send additional field infos (after house deals are finished)
  // TODO: these field infos should be the only field infos
  // currently there are still field infos in buy/sell_houses, because
  // client is used to this (every field info coupled with player info)
  for (int i=0; i<_nstreets; ++i) {
    // send update for all fields (not only these where houses have changed
    game->send_field_info(TOALL, _street[i], E_NOEVENT);
  }

  game->send_monop_info(NULL, this, E_SELL_HOUSES);

}

int Monop::sell_all_houses(Player *plr)
{
  int i;

  if (!is_yours(plr)) {
    notify_player(plr, "Game: you have no monopoly there to sell houses");
    freelog(LOG_NORMAL, "error: has no monopoly to build houses");
    return 1;
  } 

  // no tests in Street::sell_houses (see buy_houses) !

  // OK, sell houses for each street separately

  int errcode;
  for (i=0; i<_nstreets; ++i) {
    errcode = _street[i]->sell_houses(plr, _street[i]->nhouses());
//        if (errcode != 0)
//  	return errcode;
  }
  
  // send additional field infos (after house deals are finished)
  // TODO: these field infos should be the only field infos
  // currently there are still field infos in buy/sell_houses, because
  // client is used to this (every field info coupled with player info)
  for (i=0; i<_nstreets; ++i) {
    // send update for all fields (not only these where houses have changed
      game->send_field_info(TOALL, _street[i], E_NOEVENT);
  }

  game->send_monop_info(NULL, this, E_SELL_HOUSES);

  return 0;
}


/*
 * fill info packets with data
 */

void Monop::init_info_for(Player *player, struct packet_monop_init_info *pack)
{
  int i;

  pack->monopid = _id;
  pack->nstreets = _nstreets;

  for (i=0; i<_nstreets; ++i)
    pack->street[i] = _street[i]->location();
}


void Monop::info_for(Player *player, struct packet_monop_info *pack, int event)
{
  pack->monopid = _id;
  pack->event = event;

  pack->owner = (_owner != NULL) ? _owner->id() : NOBODY;
}


