// -*- C++ -*-
/***************************************************************************
                          player.cpp  -  description
                             -------------------
    begin                : Fre Jun 22 17:19:03 CEST 2001
    copyright            : (C) 2001 by Daniel Hermann
    email                : Daniel.Hermann@stud.uni-karlsruhe.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <time.h>

#include "events.h"
#include "log.h"

#include "main.h"
#include "sernet.h"

#include "plrhand.h"

#include "game.h"
#include "board.h"
#include "property.h"
#include "street.h"
#include "monop.h"
#include "player.h"
#include "cardstack.h"

static bool already_paid = false;

extern int force_end_of_sniff;

/*
 * TODO: improve freelogs (more precice information)
 */

//  #define MIN(x,y)	((x)<=(y) ? (x) : (y))

extern Game *game;

Player::Player(const char *name, const int startbudget)
  : _conn(0),
    _id(-1),	/* no game started yet */
    _location(GO),
    _nturns_in_jail(0),
    _budget(startbudget),
    _last_transaction(0),
    _last_partner(0),
    _is_alive(true),
    _ncards(0),
    _bank_debts(0),
    _ai_controlled(false)
{
  int i;

  set_name(name);
  strcpy(_addr,"---.---.---.---");

  for (i=0; i<BOARD_N_FIELDS; ++i) {
    _holdings[i] = 0;
  }

  for (i=0; i<GAME_MAX_PLAYERS; ++i) {
    _debts[i] = 0;
    _trade[i] = 0;
  }

  for (i=0; i<GAME_MAX_JAILCARDS; ++i) {
    _card[i] = 0;
  }

}
 
int const Player::bank_debts() const
{
  return _bank_debts;
}

int const Player::debts(const int plr_no) const
{
  assert(!game->nobody(plr_no));
  return _debts[plr_no];
}

int const Player::debts_sum() const
{
  int i;
  int sum = 0;

  sum += _bank_debts;
  for(i=0; i<GAME_MAX_PLAYERS; ++i) {
    sum += _debts[i];
  }

  return sum;
}

int const Player::real_holdings() const
{
  int i, sum = 0;

  //add full cash money
  sum += _budget;

  // properties (distinguish mortgaged <-> unmortgage)
  for (i=0; i<BOARD_N_FIELDS; ++i) {

    Property *prop = _holdings[i];
    Street *street = dynamic_cast<Street *>(prop);

    if (!prop) { continue; } // it's no property and/or we don't own it
    if (prop->is_mortgaged()) { // distinguish mortgaged <-> unmortgaged
      sum += prop->price() / 2;
      sum -= (int) (PROPERTY_UNMORTGAGE_FACTOR * prop->price());
    } else { // not mortgaged
      sum += prop->price();
    }
    
    // sum up halfprice of houses/hotels if it's a street (and owned by us)
    if (street) {
      sum += street->nhouses() * street->house_price() / 2;
    }

  }

  return sum;
}


int const Player::tax_holdings() const
{
  int i, sum = 0;

  // difference to real_holdings:
  //     use the full prices for mortgaged props and houses

  //add full cash money
  sum += _budget;

  // properties
  for (i=0; i<BOARD_N_FIELDS; ++i) {

    Property *prop = _holdings[i];
    Street *street = dynamic_cast<Street *>(prop);

    if (!prop) { continue; } // it's no property and/or we don't own it

    sum += prop->price();
    
    // sum up full price of houses/hotels if it's a street (and owned by us)
    if (street) {
      sum += street->nhouses() * street->house_price();
    }

  }

  return sum;
}

int const Player::liquid_holdings() const
{
  int i, sum = 0;

  //add full cash money
  sum += _budget;

  // properties (distinguish mortgaged <-> unmortgage)
  for (i=0; i<BOARD_N_FIELDS; ++i) {

    Property *prop = _holdings[i];
    Street *street = dynamic_cast<Street *>(prop);

    if (!prop) { continue; } // it's no property and/or we don't own it
    if (prop->is_mortgaged()) { // distinguish mortgaged <-> unmortgaged
      // we can't get any money out of it
    } else { // not mortgaged, we can get half price by mortgaging
      sum += prop->price() / 2;
    }
    
    // sum up halfprice of houses/hotels if it's a street (and owned by us)
    if (street) {
      sum += street->nhouses() * street->house_price() / 2;
    }

  }

  return sum;
}


void Player::connect(struct connection *conn)
{ 
  _conn = conn; 
  if (conn) {
    strcpy(_addr, conn->addr);
  } 
}

void Player::disconnect()
{
  _conn = 0;
  strcpy(_addr, "---.---.---.---");
}




void Player::cleanup_all_holdings()
{
  /* TODO: */
}



void Player::init_info_for(Player *player, 
			   struct packet_player_init_info *pack)
{
  pack->playerid = _id;
  strcpy(pack->name, _name);
  strcpy(pack->addr, _addr);
}


void Player::info_for(Player *player, struct packet_player_info *pack, 
		      int event)
{
  int i;

  pack->playerid = _id;
  pack->event = event;

  int statusflags = 0;

  statusflags |= (is_connected()) ? IS_CONNECTED : 0;
  statusflags |= (is_alive()) ? IS_ALIVE : 0;
  statusflags |= (is_afloat()) ? IS_AFLOAT : 0;
  statusflags |= (is_ai_controlled()) ? IS_AI_CONTROLLED : 0;

  pack->statusflags = statusflags;

  int actionflags = 0;

  // TODO: CAN_DECLARE_BANKRUPT ??

  pack->actionflags = actionflags;

  pack->location = _location;
  pack->nturns_in_jail = _nturns_in_jail;

  pack->budget = _budget;
  pack->transaction = _last_transaction;
  pack->partner = (_last_partner) ? _last_partner->id() : NOBODY;

  for (i=0; i<BOARD_N_FIELDS; ++i) {
    pack->holdings[i] = (_holdings[i] != 0);
  }
  pack->ncards = _ncards;

  pack->bank_debts = _bank_debts;
  for (i=0; i<GAME_MAX_PLAYERS; ++i) 
    pack->debts[i] = _debts[i];

}


/*
 * holdings
 */
void Player::holdings_add(Property *prop)
{
  _holdings[prop->location()] = prop;
}



/*
 * game flow
 */

int Player::move(int steps, int pass_go)
{ 
  int direction = (steps >= 0) ? DIRECTION_FORWARD : DIRECTION_BACKWARD;

  int newloc = (_location + steps) % BOARD_N_FIELDS;
  if (newloc < GO) newloc = BOARD_N_FIELDS + newloc;

  return move_to( newloc, pass_go, direction);
}


int Player::move_to(int newloc, int pass_go, int direction)
{
  bool passed_go = false;
  int event;

  assert(_location != JAIL);
  assert(newloc >= GO && newloc < BOARD_N_FIELDS);

  if (pass_go == PASS_GO) {	// collect money if passed GO
    if (newloc-_location < 0) {
      earn(BANK, game->income_on_GO(), E_NOEVENT);
//        notify_player(this, "You passed GO and collected $%d,", 
//  		    game->income_on_GO());
      passed_go = true;
    }
  }
  _location = newloc;

  if (direction == DIRECTION_FORWARD) {
    event = (passed_go) ? E_MOVE_PASS_GO : E_MOVE;
  } else { // DIRECTION_BACKWARD
    event = E_MOVE_BACKWARDS;
  }

  game->send_player_info(TOALL, this, event);
  
  game->board()->field(newloc)->enter(this);

  return newloc;
}

void Player::goto_jail(int event)
{
  assert(!is_in_jail());

  _location = JAIL;
  game->no_2nd_turn();
  
  game->send_player_info(TOALL, this, event);
}


int Player::leave_jail(int event)
{
  assert(is_in_jail());

  switch (event) {

  case E_LEAVE_JAIL_NORMAL:
    // leave jail in any case; if we have no money, we run into debts
    // this is to prevent players to stay in jail voluntarily
    // TODO: plr should have possibility to choose between paying fine and
    // using card *after* third dicing
    pay(BANK, GAME_JAIL_FINE, E_NOEVENT);
    break;

  case E_LEAVE_JAIL_PAY_FINE:
    if (!has_enough_money(GAME_JAIL_FINE)) {
      notify_player(this, "Game: You haven't got enough money to pay fine");
      freelog(LOG_NORMAL, "%s hasn't got enough money to pay fine", name());
      return 1;
    }

    pay(BANK, GAME_JAIL_FINE, E_NOEVENT);
    break;

  case E_LEAVE_JAIL_USE_CARD:
    if (!(_ncards>0)) {	// we have no jail cards
      notify_player(this, "Game: You haven't got any jail free cards");
      freelog(LOG_NORMAL, "%s hasn't got any jail free card", name());
      return 1;
    }
    
    _card[--_ncards]->put_back();
    delete _card[_ncards];  _card[_ncards]=0;
    break;
  }

  _location = JUST_VISITING;
  _nturns_in_jail = 0;
  game->no_2nd_turn();

  game->send_player_info(TOALL, this, event);

  return 0;
}

void Player::stay_in_jail()
{ 
  assert(is_in_jail());
  assert(_nturns_in_jail < 3);
  //  _nturns_in_jail++;	// done at next_turn in Game.cc
  game->send_player_info(TOALL, this, E_STAY_IN_JAIL);
}


void Player::add_jail_card(JailCard *card)
{ 
  _card[_ncards++] = card;
  game->send_player_info(TOALL, this, E_CARD_JAIL_FREE);
}

void Player::renovate_buildings(int pay_for_house, int pay_for_hotel)
{
  int i, nhouses = 0, nhotels = 0;

  for (i=0; i<BOARD_N_FIELDS; ++i) {
    Street *s = dynamic_cast<Street *>(_holdings[i]);
    if (s) { // it's a street and you own that street
      if (s->nhouses() <= STREET_MAX_HOUSES)
	nhouses += s->nhouses();
      else
	nhotels += 1;
    }
  }

  pay(BANK, nhouses * pay_for_house + nhotels * pay_for_hotel, 
      E_CARD_RENOVATE_BUILDINGS);

//    game->send_player_info(TOALL, this, E_CARD_RENOVATE_BUILDINGS);
  
}


void Player::update_holdings()
{
  int i;

  for (i=0; i<BOARD_N_FIELDS; ++i) {
    Property *prop = dynamic_cast<Property *>(game->board()->field(i));
    if (prop && prop->owner() == this)
      _holdings[i] = prop;
    else 
      _holdings[i] = 0;
  }
}

void Player::go_bankrupt(Player *dest)
{
  assert(is_bankrupt());

  int i;

  // first we get money by selling all the houses we have
  for (i=0; i<BOARD_N_FIELDS; ++i) {
    Property *prop = _holdings[i];
    Street *street = dynamic_cast<Street *>(prop);
    
    if (prop) { // field is property and we own it

      // sell houses/hotels to the bank
      if (street && street->monop()->is_monop()) { 
	// property is street and belongs to established monopoly
	street->monop()->sell_all_houses(this);
      }

    }

  }

  // next we clear the debts to the creditor ...
  if (dest) _debts[dest->id()] = 0;
  else _bank_debts = 0;

  // and transact cash (after we got all money from selling houses etc.)
  // as much as we have, could be less than the debts !
  pay(dest, _budget, E_NOEVENT);


  // now the creditor gets all the properties (probably he has to pay for
  // mortgaged ones)
  for (i=0; i<BOARD_N_FIELDS; ++i) {
    Property *prop = _holdings[i];
    
    if (prop) { // field is property and we own it
      prop->change_owner(dest);
    }

  }

  update_holdings();
  if (dest) dest->update_holdings();


  game->send_player_info(TOALL, this, E_GO_BANKRUPT);
  if (dest) game->send_player_info(TOALL, dest, E_BANKRUPT_CREDITOR);
  // TODO: also send some E_BANKRUPT_CREDITOR info when dest == BANK

//    game->send_player_info(TOALL, this, E_GO_BANKRUPT);
//    game->send_player_info(TOALL, dest, E_BANKRUPT_CREDITOR);

  freelog(LOG_NORMAL, "bankruptcy of %s finished (creditor %s)", 
	  this->name(), (dest)? dest->name() : "bank");

}

void Player::die()
{
  // we die
  _is_alive = false;
  
  game->send_player_info(TOALL, this, E_DIE);

  freelog(LOG_NORMAL, "player %s dies", 
	  this->name());
}

/*
 * dest can be 0, means bank
 */
void Player::pay(Player *dest, int money, int event)
{
  assert(money >= 0);
  bool was_afloat = is_afloat();

  // for cards we pay what we have and don't run into debts
  if (event == E_CARD_PAY) {
    if (has_enough_money(money)) {
      _budget -= money;
      _last_transaction = - money;
    } else { // we pay all we have
      _budget = 0;
    }
//      game->send_player_info(TOALL, this, event);
    return;
  }

  if(event == E_CARD_PAY_TO_ALL) {
    // we are sure we have the money, see pay_to_all()
    _budget -= money;
    _last_transaction = - money;
//      game->send_player_info(TOALL, this, event);
    return;
  }


  // we also run into new debts when we already are in debts to another player
  // no matter if we can pay these new debts
  if (has_enough_money(money) && was_afloat) { // we can pay immediately
    _budget -= money;
    _last_transaction = - money;
    if (dest) { // pay to another player
      already_paid = true;
      dest->earn(this, money, game->complement_event(event));
    }
  } else { // we run into debts
    if (dest) { // debts to another player
      _debts[dest->id()] += money;
    } else { // debts to the bank
      _bank_debts += money;
    }
    _last_transaction = - money;

    // rewrite the event to ...SHOULD...
    switch (event) {
    case E_PAY_RENT: event = E_SHOULD_PAY_RENT; break;
    case E_PAY_TAX: event = E_SHOULD_PAY_TAX; break;
    case E_CARD_PAY: event = E_CARD_SHOULD_PAY; break;
    case E_CARD_RENOVATE_BUILDINGS: 
      event = E_CARD_SHOULD_RENOVATE_BUILDINGS;
      break;
    default: break; // leave event unchanged by default
    }

    // TODO: this notification should come after the other infos
    if (was_afloat)
      notify_player(this, "Game: You run into debts!");
    else
      notify_player(this, "Game: You are in debt!");
    notify_player(this, "Game: Settle them by mortgaging/selling/trading");
    freelog(LOG_NORMAL, "%s runs into debts/is in debt", name());
  }

  // TODO: E_RECV comes before E_PAY (inconsistent, but what's the problem?)
  if (event != E_NOEVENT) { 
    game->send_player_info(TOALL, this, event);
  }

}

void Player::earn(Player *src, int money, int event)
{
  _budget += money;
  _last_transaction = money;

  // if we receive something from src, he should have paid it
  assert(src == BANK || already_paid);	
  already_paid = false;

  if (event != E_NOEVENT) { 
    game->send_player_info(TOALL, this, event);
  }

//    // try to settle debts (in case we have some)
//    // this is not a good idea to do it automatically
//    // because debts can happen when players don't have their turn!
//    // -> commented out
//    if (debts_sum() > 0 && settle_debts() == 0) { 
//      // we had debts and could settle them
//      // some notification needed here ??
//      freelog(LOG_NORMAL, "%s successfully settled debts", src->);
//    }
    
    
//    if (event == E_CARD_COLLECT || event == E_CARD_COLLECT_FROM_ALL)
//      game->send_player_info(TOALL, this, event);
  
}

void Player::pay_to_all(int money, int event)
{
  assert(money >= 0);

  int i;
  // only pay what we have as cash
  // we don't get bankrupt by pay_to_all !
  int money_affordable = MIN(money, _budget / (game->nplayers()-1));
  int total_money = money_affordable * (game->nplayers()-1);
  
  for (i=0; i<game->nplayers(); ++i)
    if (i != _id) {	// all players earn from bank
      // we use E_NOEVENT here, because client reacts on last pay at end of
      // this method
      game->player(i)->earn(BANK, money_affordable, E_NOEVENT);
      game->send_player_info(TOALL, game->player(i), E_NOEVENT);
    }
  // I pay all the amount to bank
  pay(BANK, total_money, event);

  if (money > money_affordable) { // we have paid less than requested
    notify_player(this, ( "Game: You are lucky, you only paid %d%% "
			  "of the requested amount" ), 
		  money_affordable * 100 / money);
    freelog(LOG_NORMAL, "%s paid only %d%% of the requested amount", 
	    name(), (money_affordable * 100) / money);
  }

}

void Player::earn_from_all(int money, int event)
{
  assert(money >= 0);

  int i;
  int total_money = money * (game->nplayers()-1);
  
  for (i=0; i<game->nplayers(); ++i)
    if (i != _id) {	// all players pay to bank
      // we use E_NOEVENT here, because client reacts on last earn at end of
      // this method
      game->player(i)->pay(BANK, money, E_NOEVENT);
      game->send_player_info(TOALL, game->player(i), E_NOEVENT);
    }
  // I get all back from bank
  earn(BANK, total_money, event);
  
}


int Player::settle_debts()
{
  int i;

  if (!is_afloat()) { // plr still can't settle debts
//      notify_player(this, ("Game: You still don't have enough money "
//  			 "to settle your debts"));
//      notify_player(this, "Game: get more money or declare yourself bankrupt");
    freelog(LOG_NORMAL, "%s can't settle debts, still isn't afloat", name());
    return 1;
  }

  // if we have no debts, just return success
  if (debts_sum() == 0) return 0;

  // we use E_NOEVENT in the following payments, because client reacts on last
  // player_info at end of this method
  
  // settle debts to bank
  if (_bank_debts > 0)
    pay(BANK, _bank_debts, E_NOEVENT);
  _bank_debts = 0;

  // settle your debts to each player
  for (i=0; i<game->nplayers(); ++i) {
    if (_debts[i] > 0) {
      pay(game->player(i), _debts[i], E_NOEVENT);
      _debts[i] = 0;
      game->send_player_info(TOALL, game->player(i), E_NOEVENT);
    }
  }
  
  game->send_player_info(TOALL, this, E_SETTLE_DEBTS);

  return 0; // settle debts was successful

}


void Player::set_ai_controlled(bool ai_controlled)
{
  _ai_controlled = ai_controlled;
}

void Player::ai_wait(int seconds_min)
{
  force_end_of_sniff=0;
  freelog(LOG_DEBUG, "sniffingpackets in ai_wait");

  // wait until all animations are finished
  while( game->anim_in_progress() ) {
    sniff_packets(time(NULL)+1);  // ignore return code of sniff_packets();
  }
  
  time_t t_start = time(NULL);
  time_t t_end_min = t_start+seconds_min;

  // wait another seconds_min seconds
  while (time(NULL) < t_end_min ) {
    sniff_packets(t_end_min);  // ignore return code of sniff_packets()
  }

}

void Player::ai_roll_dice()
{
  freelog(LOG_VERBOSE, "AI: %s rolls dice", name());

  struct packet_generic_integer packet;
  packet.value=0;
  game->handle_dice_roll_request(this, &packet);
}

void Player::ai_settle_debts()
{
  if (debts_sum() == 0) return;

  freelog(LOG_VERBOSE, "AI: %s tries to settle debts", name());
  
  int i;

  struct packet_generic_integer packet;

  if (is_bankrupt()) {
    freelog(LOG_VERBOSE, "AI: %s is bankrupt, give up", name());
    game->handle_declare_bankrupt(this);
    return;
  }
    
  while(1) {  // pseudo loop

    freelog(LOG_VERBOSE, "AI: %s tries to mortgage properties", name());

    // mortgage properties not part of a monopoly (largest ones first)
    i=BOARD_N_FIELDS-1;
    while (!is_afloat() && i>=0) {
      Property *prop = _holdings[i];
      Street *street = dynamic_cast<Street *>(prop);

      if ( (prop && !street) ||
	   (street && !street->monop()->is_monop()) ) {
	packet.value = i;
	game->handle_property_mortgage(this, &packet);
      }
      i--;
    }

    // mortgage properties of monopolies without houses (largest ones first)
    i=BOARD_N_FIELDS-1;
    while (!is_afloat() && i>=0) {
      Property *prop = _holdings[i];
      Street *street = dynamic_cast<Street *>(prop);

      if ( street && !street->monop()->has_houses() ) {
	packet.value = i;
	game->handle_property_mortgage(this, &packet);
      }
      i--;
    }

    if (is_afloat()) break;

    freelog(LOG_VERBOSE, "AI: %s tries to sell houses", name());

    // sell houses on monopolies and mortgage the streets (largest ones first)

    i=BOARD_N_FIELDS-1;
    while (!is_afloat() && i>=0) {
      Property *prop = _holdings[i];
      Street *street = dynamic_cast<Street *>(prop);

      if ( street && street->monop()->is_monop() ) {
	// sell the houses ...
	while(street->monop()->has_houses()) {
	  street->monop()->sell_one_house(this);
	  if (is_afloat()) break;
	}
	
	// ... and mortgage the street (if necessary)
	if (!is_afloat()) {	
	  packet.value = i;
	  game->handle_property_mortgage(this, &packet);
	}
      }
      i--; // next street
    }

    if (is_afloat()) {
      break;
    } else {
      // we should never get here!
      if (!is_bankrupt()) {
	freelog(LOG_FATAL, "AI: %s has entered undefined state in "
		"ai_settle_debts()!!", name());
      }
    }

    // \todo: settle my debts by
    //   1.) selling houses and hotels
    //   2.) mortgaging properties
    //   3.) trading   <-- TODO
    //   4.) declaring myself bankrupt
    // \todo

  }

  freelog(LOG_VERBOSE, "AI: %s has prepared to settled debts", name());
  this->settle_debts();

}

void Player::ai_leave_jail()
{
  if (!is_in_jail()) return;

  freelog(LOG_VERBOSE, "AI: %s considers to leave jail", name());

  struct packet_generic_integer packet;

  if (ncards()>0) {
    packet.value = E_LEAVE_JAIL_USE_CARD;
  } else if (has_enough_money(GAME_JAIL_FINE)) {
    packet.value = E_LEAVE_JAIL_PAY_FINE;
  } else {
    // stay in jail
    return;
  }

  freelog(LOG_VERBOSE, "AI: %s leaves jail", name());
  game->handle_jail_free_request(this, &packet);

}

void Player::ai_manage_holdings()
{
  ai_buy_property();
  ai_unmortgage();
  ai_buy_houses();
}

void Player::ai_end_turn()
{
  freelog(LOG_VERBOSE, "AI: %s ends turn", name());
  game->handle_turn_done(this);
}


void Player::ai_buy_property()
{
  if (location() == JAIL) return;

  Property *prop = dynamic_cast<Property *>(game->board()->field(location()));
  Street *street = dynamic_cast<Street *>(prop);
    
  struct packet_generic_integer packet;
  packet.value = location();

  if (!prop || prop->is_owned()) {  
    // no property or property is already owned
    return;
  }

  freelog(LOG_VERBOSE, "AI: %s considers to buy property", name());

  if (!has_enough_money(prop->price())) {
    // \todo: we may want to mortgage/trade other stuff to get this property
    //        in case that
    //   1.) the property is important to get a monopoly
    //   2.) the property can prevent a monopoly (large chance for trade)
    return;
  }

  // alright, let's buy it
  freelog(LOG_VERBOSE, "AI: %s buys property", name());
  game->handle_property_buy(this, &packet);

  ai_wait(1);
}

void Player::ai_unmortgage()
{
  // unmortgage highest monopoly (\todo: preferred)
  int m=BOARD_N_MONOPS-1;
  while (m>=0) {
    Monop *monop = game->board()->monop(m);
    if ( monop->is_yours(this) ) {
      // unmortgage all streets in monopoly
      for (int s=monop->nstreets()-1; s>=0; --s) {
	if ( monop->street(s)->is_mortgaged() &&
	     budget() > monop->street(s)->unmortgage_price() ) {
	  struct packet_generic_integer packet;
	  freelog(LOG_VERBOSE, "AI: %s considers to unmortgage "
		  "some property", name());
	  packet.value = monop->street(s)->location();
	  game->handle_property_unmortgage(this, &packet);
	}
      }
      break;
    }
    m--;
  }
}

void Player::ai_buy_houses()
{
  struct packet_house_request packet;

  // build on highest monopoly (\todo: the preferred monopoly)
  int m=BOARD_N_MONOPS-1;
  while (m>=0) {

    Monop *monop = game->board()->monop(m);
    if ( monop->is_yours(this) ) {

      // we assume here that house cost is equal on all streets in monopoly
      int house_price = monop->street(0)->house_price();

      // add as many houses as the budget allows
      int nhouses_to_add = budget() / house_price;
      while(nhouses_to_add>0) {

	packet.monopid = m;
	for (int s=0; s<MAX_STREETS_IN_MONOP; ++s) packet.nhouses[s]=0;

	// find street with least houses ...
	int least_houses=STREET_MAX_HOUSES + 1 + 1; // larger than maximum
	int street_least_houses;
	bool has_mortgaged_streets = false;
	for (int s=0; s<monop->nstreets(); ++s) {
	  if (monop->street(s)->is_mortgaged()) has_mortgaged_streets=true;
	  if (monop->street(s)->nhouses() <= least_houses) {
	    least_houses = monop->street(s)->nhouses();
	    street_least_houses = s;
	  }
	}
	if (has_mortgaged_streets) break;
	// ... and add houses to that street
	if ( least_houses < STREET_MAX_HOUSES + 1 ) {
	  packet.nhouses[street_least_houses] = 1;
	  freelog(LOG_VERBOSE, "AI: %s considers to build a house in "
		  "monopoly %d (street %d)", name(), m, street_least_houses);
	  game->handle_houses_buy(this, &packet);
	}
	if ( monop->street(street_least_houses)->nhouses() == least_houses ) {
	  // there was a problem or unusual condition building the house
	  freelog(LOG_VERBOSE, "AI: %s didn't get the new house, stops trying",
		  name());
	  break;
	}
	nhouses_to_add--;
    
      }

    }

    m--;
  }

}



