// -*- C++ -*-
/***************************************************************************
                          player.h  -  description
                             -------------------
    begin                : Fre Jun 22 17:19:03 CEST 2001
    copyright            : (C) 2001 by Daniel Hermann
    email                : Daniel.Hermann@stud.uni-karlsruhe.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef PLAYER_H
#define PLAYER_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>

#include <string.h>

#include "global.h"
#include "packets.h"

#include "board.h"

class Trade;

enum player_statusflags {
  IS_CONNECTED = 1,
  IS_ALIVE = 2,
  IS_AFLOAT = 4,
  IS_AI_CONTROLLED = 8
};

enum player_actionflags {
  CAN_DECLARE_BANKRUPT = 1
};

enum {
  PASS_GO,
  DONT_PASS_GO,
  DIRECTION_FORWARD,
  DIRECTION_BACKWARD
};


class Property;
class JailCard;


class Player {
  
public:


public:
  
  /*
   * server constructor
   */
  Player(const char *name, const int startbudget);
  
  /*
   * client constructor
   */
  Player(struct packet_player_init_info *);


  const char * name() const
    { return _name; }
  // to avoid warnings in C-code, return non-const connection
  struct connection *conn() const 
    { return _conn; }
  const char * addr() const 
    { return _addr; }

  const bool is_connected() const	  // don't use this on client side !
    { return (_conn) ? true : false; }

  const int id() const
    { return _id; }

  const int location() const
    { return _location; }
  const int nturns_in_jail() const
    { return _nturns_in_jail; }
  const bool is_in_jail() const
    { return _location == JAIL; }

  const int budget() const
    { return _budget; }

  int const bank_debts() const;
  int const debts(const int plr_no) const;
  int const debts_sum() const;
  
  int const real_holdings() const;
  int const tax_holdings() const;
  int const liquid_holdings() const;

  const int last_transaction() const
    { return _last_transaction; }
  Player * const last_partner() const
    { return _last_partner; }

  const bool is_alive() const
    { return _is_alive; }
  const bool is_afloat() const
    { return has_enough_money(debts_sum()); }
  const bool is_bankrupt() const
    { return liquid_holdings() - debts_sum() < 0; }

  Property * const holdings(const int i) const
    { return _holdings[i]; }

  const int nrailroads() const
    { return ( ( (_holdings[RRSOUTH])  ? 1 : 0 ) +
	       ( (_holdings[RRWEST])   ? 1 : 0 ) +
	       ( (_holdings[RRNORTH])  ? 1 : 0 ) +
	       ( (_holdings[RRCENTRAL])? 1 : 0 )  ); }

  const int nutils() const
    { return ( ( (_holdings[UELECTRIC]) ? 1 : 0 ) +
	       ( (_holdings[UWATER])    ? 1 : 0 )  ); }
  const int ncards() const
    { return _ncards; }

  const bool has_enough_money(const int price) const
    { return (_budget>=price); }

  void set_name(const char *name) 
    { strncpy(_name, name, PLAYER_MAX_LENGTH_NAME); }
  void set_id(const int id)
    { _id = id; }
  
  /*
   * holdings
   */
  void holdings_add(Property *);

  void connect(struct connection *conn);
  void disconnect();

  void cleanup_all_holdings();

  /*
   * generate infos (server side)
   */
  void init_info_for(Player *, struct packet_player_init_info *);
  void info_for(Player *, struct packet_player_info *, int event);
  
  /*
   * update data with infos from server (on client side)
   */
  
  void update(struct packet_player_info *);


  /*
   * movement actions
   */

  int move(int steps, int pass_go);

  int move_to(int newloc, int pass_go, int direction = DIRECTION_FORWARD);

  /*
   * jail
   */
  void goto_jail(int event);
  int leave_jail(int event);
  void stay_in_jail();
  void inc_nturns_in_jail()
    { 
      assert(_nturns_in_jail <= 4);	// in 4th turn, player gets free
      _nturns_in_jail++;
    }
  void add_jail_card(JailCard *card);
  
  void renovate_buildings(int pay_for_house, int pay_for_hotel);

  void update_holdings();

  void go_bankrupt(Player *dest);
  void die();

  /*
   * money transactions
   */
  void pay(Player *dest, int money, int event);
  void earn(Player *src, int money, int event);

  void pay_to_all(int money, int event);
  void earn_from_all(int money, int event);

  int settle_debts();

  /*
   * trade affairs
   */
  Trade *get_trade_with(Player *plr)
    { return _trade[plr->id()]; }
  void set_trade_with(Player *plr, Trade *trade)
    { _trade[plr->id()] = trade; }
  void clear_trade_with(Player *plr)
    { _trade[plr->id()] = 0; }


  //
  // AI methods
  //

  bool is_ai_controlled() { return _ai_controlled; }

  void set_ai_controlled(bool ai_controlled);

  void ai_wait(int seconds_min);
  void ai_roll_dice();
  void ai_settle_debts();
  void ai_leave_jail();
  void ai_manage_holdings();
  void ai_end_turn();

  // sub-methods
  void ai_buy_property();
  void ai_unmortgage();
  void ai_buy_houses();

protected:

  char _name[PLAYER_MAX_LENGTH_NAME];

  // network stuff
  struct connection *_conn;
  char _addr[PLAYER_MAX_LENGTH_ADDRESS];

  int _id;
  int _location;
  int _nturns_in_jail;

  int _budget;
  int _last_transaction;
  Player *_last_partner;

  bool _is_alive;

  // note: properties the player doesn't own are 0
  Property * _holdings[BOARD_N_FIELDS];

  JailCard *_card[GAME_MAX_JAILCARDS];
  int _ncards;

  // debts
  int _bank_debts;
  int _debts[GAME_MAX_PLAYERS];

  Trade *_trade[GAME_MAX_PLAYERS];

  // note: trades that are not existent are 0

  //
  // AI attributes
  //

  bool _ai_controlled;
  
};


#define players_iterate(PI_player)                                            \
{                                                                             \
  Player *PI_player;                                                          \
  int PI_p_itr;                                                               \
  for (PI_p_itr = 0; PI_p_itr < game->nplayers(); PI_p_itr++) {               \
    PI_player = game->player(PI_p_itr);

#define players_iterate_end                                                   \
  }                                                                           \
}


#endif /* PLAYER_H */
