// -*- C++ -*-
/***************************************************************************
                          property.cpp  -  description
                             -------------------
    begin                : Fre Jun 22 17:19:03 CEST 2001
    copyright            : (C) 2001 by Daniel Hermann
    email                : Daniel.Hermann@stud.uni-karlsruhe.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <assert.h>

#include "global.h"
#include "log.h"
#include "events.h"
#include "console.h"

#include "plrhand.h"

#include "game.h"
#include "player.h"

#include "property.h"

extern Game *game;


Property::Property(const int location, const int price)
  : Field(location),
    _owner(BANK),
    _price(price),
    _is_mortgaged(false)
{
  // _type is not set, there are no pure properties, only derived ones
}

Property::~Property()
{
}

void Property::enter(Player *player)
{
  freelog(LOG_NORMAL, "Player has entered property %d", _location);

  game->send_field_info(player, this, E_ENTER_FIELD);

  // pay rent if property is owned by another player and not mortgaged
  if ( is_owned() && !is_yours(player) && !is_mortgaged() ) {
    pay_rent(player);
  } 
  
}


void Property::pay_rent(Player *plr)
{
  assert(is_owned());
  int xrent = rent();

  if (!plr->has_enough_money(xrent)) {
    notify_player(plr, ("Game: Attention: You haven't got enough money to pay "
			"rent"));
    freelog(LOG_NORMAL, "warning: %s has not enough money to pay rent", 
	    plr->name());
  }
    
  plr->pay(owner(), xrent, E_PAY_RENT);

//    game->send_player_info(TOALL, plr, E_PAY_RENT);
//    game->send_player_info(TOALL, owner(), E_RECV_RENT);
  
  // TODO: in case of debts, use "should pay" here
  freelog(LOG_NORMAL, "%s paid a rent of $%d to %s", 
	  plr->name(), xrent, owner()->name());

}

int Property::buy(Player *plr)
{
  if (is_owned()) {
    notify_player(plr, "Game: property is owned, you can't buy it");
    freelog(LOG_NORMAL, "error: property is already owned");
    return 1;
  } else if (!plr->has_enough_money(_price)) {
    notify_player(plr, 
		  "Game: You haven't got enough money to buy this property");
    freelog(LOG_NORMAL, "error: hasn't got enough money to buy property");
    return 1;
  }

  _owner = plr;
  plr->pay(BANK, _price, E_NOEVENT);
  plr->holdings_add(this);

  game->send_player_info(TOALL, plr, E_BUY_PROP);
  game->send_field_info(TOALL, this, E_BUY_PROP);

  return 0;
}


int Property::mortgage(Player *plr)
{
  if (!is_yours(plr)) {
    notify_player(plr, "Game: property is not yours, you can't mortgage it");
    freelog(LOG_NORMAL, "error: property is not owned by him");
    return 1;
  } else if (_is_mortgaged) {	// property already mortgaged
    notify_player(plr, "Game: property is already mortgaged");
    freelog(LOG_NORMAL, "error: property is already mortgaged");
    return 1;
  }

  _is_mortgaged = true;
  plr->earn(BANK, _price/2, E_MORTGAGE_PROP);

//    game->send_player_info(TOALL, plr, E_MORTGAGE_PROP);
  game->send_field_info(TOALL, this, E_MORTGAGE_PROP);
  
  return 0;
}

int Property::unmortgage(Player *plr)
{
  int money_to_pay = _price/2 + (int) (PROPERTY_UNMORTGAGE_FACTOR * _price);

  if (!is_yours(plr)) {
    notify_player(plr, "Game: property is not yours, you can't unmortgage it");
    freelog(LOG_NORMAL, "error: property is not yours");
    return 1;
  } else if (!_is_mortgaged) {	// property isn't mortgaged
    notify_player(plr, "Game: property isn't mortgaged");
    freelog(LOG_NORMAL, "error: property isn't mortgaged");
    return 1;
  } else if (!plr->has_enough_money(money_to_pay)) {
    notify_player(plr, "Game: You haven't got enough money to unmortgage "
		  "this property");
    freelog(LOG_NORMAL, ("error: hasn't got enough money to unmortgage "
			 "property"));
    return 1;
  }

  _is_mortgaged = false;
  plr->pay(BANK, money_to_pay, E_UNMORTGAGE_PROP);

//    game->send_player_info(TOALL, plr, E_UNMORTGAGE_PROP);
  game->send_field_info(TOALL, this, E_UNMORTGAGE_PROP);
  
  return 0;
}


int Property::change_owner(Player *new_owner)
{
//    assert(new_owner);
  // new_owner == 0 means the current owner gives the property back to the bank

  // return success and do nothing in case nothing has changed
  if (new_owner == _owner)
    return 0;

  _owner = new_owner;

  if (new_owner) { // new_owner is not bank
    // if mortgaged, new owner must pay an interest of 10 percent
    if (_is_mortgaged) {
      int interest = (int) (PROPERTY_UNMORTGAGE_FACTOR * _price);
      _owner->pay(BANK, interest, E_NOEVENT);
      freelog(LOG_NORMAL, "%s paid interest: $%d",
	      new_owner->name(), interest);
      // TODO: make sure a send_player_info occurs after this
      // -> new flag in class Player: plr_changed = true ???
    }
  } else { // new_owner is BANK, so remove some status flags
    _is_mortgaged = false;
  }
  
  game->send_field_info(TOALL, this, E_OWNER_CHANGED);

  return 0;
}


int const Property::status(const Player *player) const
{
  int statusflags = 0;

  statusflags |= (_owner != BANK) ? IS_OWNED : 0;
  statusflags |= (_owner == player) ? IS_YOURS : 0;
  statusflags |= (_is_mortgaged) ? IS_MORTGAGED : 0;

  return statusflags;
}

int const Property::actions(const Player *player) const
{
  int actionflags = 0;

//    bool can_buy = (_owner == BANK) && player->has_enough_money(_price);
  bool can_buy = (_owner == BANK) && (player->location() == _location);
  bool can_mortgage = ( (_owner == player) && !_is_mortgaged );
  bool can_unmortgage = ( (_owner == player) 
			  && _is_mortgaged );
//  			  && player->has_enough_money(_price) );

  actionflags |= (can_buy) ? CAN_BUY : 0;
  actionflags |= (can_mortgage) ? CAN_MORTGAGE : 0;
  actionflags |= (can_unmortgage) ? CAN_UNMORTGAGE : 0;

  return actionflags;
}


void Property::info_for(const Player *player, struct packet_field_info *pack, 
			int event)
{
  pack->location = _location;
  pack->event = event;

  pack->statusflags = status(player);
  pack->actionflags = actions(player);

  pack->owner = (_owner != BANK) ? _owner->id() : NOBODY;
  pack->nhouses = 0;

}

