// -*- C++ -*-
/***************************************************************************
                          street.cpp  -  description
                             -------------------
    begin                : Fre Jun 22 17:19:03 CEST 2001
    copyright            : (C) 2001 by Daniel Hermann
    email                : Daniel.Hermann@stud.uni-karlsruhe.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <assert.h>

#include "log.h"
#include "events.h"

#include "plrhand.h"

#include "game.h"
#include "player.h"
#include "monop.h"

#include "street.h"

extern Game *game;

Street::Street(const int location, 
	       const int price, const int house_price,
	       const int rent[])
  : Property(location, price),
    _monop(NULL),
    _house_price(house_price),
    _nhouses(0),
    _is_monop(false)
{
  int i;

  _type = STREET;

  for (i=0; i<STREET_N_RENT_CLASSES; ++i) {
    _rent[i] = rent[i];
  }
}

int Street::buy(Player *plr)
{
  if (is_owned()) {
    notify_player(plr, "Game: street is owned, you can't buy it");
    freelog(LOG_NORMAL, "error: street is already owned");
    return 1;
  } else if (!plr->has_enough_money(_price)) {
    notify_player(plr, 
		  "Game: You haven't got enough money to buy this street");
    freelog(LOG_NORMAL, "error: hasn't got enough money to buy street");
    return 1;
  }

  _owner = plr;
  plr->pay(BANK, _price, E_NOEVENT);
  plr->holdings_add(this);

  game->send_player_info(TOALL, plr, E_BUY_PROP);
  game->send_field_info(TOALL, this, E_BUY_PROP);

  _monop->check();

  return 0;
}


int Street::mortgage(Player *plr)
{
  if (!is_yours(plr)) {
    notify_player(plr, "Game: street is not yours, you can't mortgage it");
    freelog(LOG_NORMAL, "error: street is not his");
    return 1;
  } else if (_is_mortgaged) {	// street already mortgaged
    notify_player(plr, "Game: street is already mortgaged");
    freelog(LOG_NORMAL, "error: street is already mortgaged");
    return 1;
  } else if (_nhouses > 0) {	// can't mortgage street with houses
    notify_player(plr, "Game: street with houses can't be mortgaged");
    freelog(LOG_NORMAL, "error: street with houses can't be mortgaged");
    return 1;
  } else if (!_monop->no_houses()) { // streets within monop have houses
    notify_player(plr, ("Game: there are houses on that monopoly, "
			"you can't mortgage"));
    freelog(LOG_NORMAL, "error: houses in monopoly, mortgaging not possible");
    return 1;
  }

  _is_mortgaged = true;
  plr->earn(BANK, _price/2, E_MORTGAGE_PROP);

//    game->send_player_info(TOALL, plr, E_MORTGAGE_PROP);
  game->send_field_info(TOALL, this, E_MORTGAGE_PROP);
  
  return 0;
}

int Street::change_owner(Player *new_owner)
{
//    assert(new_owner);
  // new_owner == 0 means the current owner gives the property back to the bank

  // return success and do nothing in case nothing has changed
  if (new_owner == _owner)
    return 0;

  if (_nhouses > 0) {	// can't change owner of street with houses
    freelog(LOG_NORMAL, "error: can't change owner of street with houses");
    return 2;
  }

  freelog(LOG_NORMAL, "street %d changed owner: old=%s, new=%s", 
	  _location, _owner->name(), new_owner->name());

  _owner = new_owner;

  if (new_owner) { // new_owner is not bank
    // if mortgaged, new owner must pay an interest of 10 percent
    if (_is_mortgaged) {
      int interest = (int) (PROPERTY_UNMORTGAGE_FACTOR * _price);
      _owner->pay(BANK, interest, E_NOEVENT);
      freelog(LOG_NORMAL, "%s paid interest: $%d",
	      new_owner->name(), interest);
      // TODO: make sure a send_player_info occurs after this
      // -> new flag in class Player: plr_changed = true ???
    }
  } else { // new_owner is BANK, so remove some status flags
    _is_mortgaged = false;
    // monopoly related issues are managed by Monop::check() later
  }

  game->send_field_info(TOALL, this, E_OWNER_CHANGED);

  _monop->check();

  return 0;
}


int Street::buy_houses(Player *plr, const int addhouses)
{
  assert(plr);
  // these are tested in Monop::buy_houses
  assert(_is_monop);
  assert(_nhouses + addhouses <= STREET_MAX_HOUSES + 1);
  
  _nhouses += addhouses;
  plr->pay(BANK, addhouses * _house_price, E_BUY_HOUSES);

//    game->send_player_info(TOALL, plr, E_BUY_HOUSES);
  game->send_field_info(TOALL, this, E_BUY_HOUSES);

  freelog(LOG_NORMAL, "%d houses on street %d bought", addhouses, _location);

  return 0;
}




int Street::sell_houses(Player *plr, const int subhouses)
{
  assert(plr);
  // these are tested in Monop::sell_houses
  assert(_is_monop);
  assert(_nhouses - subhouses >= 0);
  
  _nhouses -= subhouses;
  plr->earn(BANK, subhouses * _house_price/2, E_SELL_HOUSES);

//    game->send_player_info(TOALL, plr, E_SELL_HOUSES);
  game->send_field_info(TOALL, this, E_SELL_HOUSES);

  freelog(LOG_NORMAL, "%d houses on street %d sold", subhouses, _location);

  return 0;
}


int const Street::status(const Player *player) const
{
  int statusflags = 0;

  statusflags |= (_owner != BANK) ? IS_OWNED : 0;
  statusflags |= (_owner == player) ? IS_YOURS : 0;
  statusflags |= (_is_mortgaged) ? IS_MORTGAGED : 0;
  statusflags |= (_is_monop) ? IS_MONOP : 0;

  return statusflags;
}

int const Street::actions(const Player *player) const
{
  int actionflags = 0;

//    bool can_buy = (_owner == BANK) && player->has_enough_money(_price);
  bool can_buy = (_owner == BANK) && (player->location() == _location);
  bool can_mortgage = ( (_owner == player) 
			&& !_is_mortgaged
			&& _monop->no_houses() );
  bool can_unmortgage = ( (_owner == player)
			  && _is_mortgaged );
//  			  && player->has_enough_money(_price) );
  bool can_build_houses = ( (_owner == player) && _is_monop 
			    && _nhouses < STREET_MAX_HOUSES + 1 );
//  			    && player->has_enough_money(house_price()) );
  bool can_sell_houses = ( (_owner == player) && _nhouses > 0 );

  actionflags |= (can_buy) ? CAN_BUY : 0;
  actionflags |= (can_mortgage) ? CAN_MORTGAGE : 0;
  actionflags |= (can_unmortgage) ? CAN_UNMORTGAGE : 0;
  actionflags |= (can_build_houses) ? CAN_BUILD_HOUSES : 0;
  actionflags |= (can_sell_houses) ? CAN_SELL_HOUSES : 0;

  return actionflags;
}


void Street::init_info_for(const Player *player,
			   struct packet_field_init_info *pack)
{
  int i;

  pack->location = _location;
  pack->type = _type;
  pack->price = _price;
  pack->house_price = _house_price;
  for (i=0; i<STREET_N_RENT_CLASSES; ++i)
    pack->rent[i] = _rent[i];
}

void Street::info_for(const Player *player, struct packet_field_info *pack, 
		      int event)
{
  pack->location = _location;
  pack->event = event;

  pack->statusflags = status(player);
  pack->actionflags = actions(player);

  pack->owner = (_owner != BANK) ? _owner->id() : NOBODY;
  pack->nhouses = _nhouses;
    
}


