// -*- C++ -*-
/***************************************************************************
                          trade.cpp  -  description
                             -------------------
    begin                : Fre Jun 22 17:19:03 CEST 2001
    copyright            : (C) 2001 by Daniel Hermann
    email                : Daniel.Hermann@stud.uni-karlsruhe.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include "global.h"
#include "log.h"
#include "events.h"

#include "plrhand.h"

#include "game.h"
#include "player.h"
#include "field.h"
#include "property.h"
#include "street.h"
#include "monop.h"

#include "trade.h"

// make it more obvious how "IS_YOURS in prop arrays is interpreted

#define IS_SRC	IS_YOURS

/*
 * TODO: make prop-arrays easier, only one value 
 * (to decide, who offers which property, we can use the owner of the property)
 */


int Trade::count = 0;

extern Game *game;

Trade::Trade(Player *src, Player *dest, int money_src, int prop_src[])
  : _id(count++),
    srcplr(src),
    destplr(dest),
    _srcmoney(money_src),
    _destmoney(0),
    _last_offer(T_SRC),
    _dest_replied_once(false),
    _finishedflag(T_NOT_FINISHED)
{
  assert(src && dest);

  int i;

  // in prop we store T_NOTRADE or the statusflag. This includes the owner by 
  // interpreting the IS_YOURS flag as IS_SRC !

  // we allow the offer to be invalid here, which is tested only in agree
  // so we simply copy prop_src into _srcprop
  // TODO: generate warnings
  for(i=0; i<BOARD_N_FIELDS; ++i) {
    _srcprop[i] = prop_src[i];
    // initialize destprop already here
    _destprop[i] = T_NOTRADE;
  }

  // notify players

//    notify_player(destplr, "Game: %s wants to trade with you", srcplrname());

  freelog(LOG_NORMAL, "new trade %d (%s to %s) initiated", 
	  _id, srcplrname(), destplrname());

  game->send_trade(NULL, this, E_TRADE_OFFER, T_SRC);

  test_money_in_offer(srcplr);

}


int Trade::reply(Player *plr, int money, int prop[])
{
  int i;

  assert(plr == srcplr || plr == destplr);

  // who replies
  bool is_src = (plr == srcplr);
  Player *other = (is_src)?destplr:srcplr;
  int *propptr = (is_src) ? _srcprop : _destprop;
  int *moneyptr = (is_src) ? &_srcmoney : &_destmoney;

  // we allow the offer to be invalid here, which is tested only in agree
  // so we simply copy prop into _srcprop/_destprop
  // TODO: generate warnings
  for(i=0; i<BOARD_N_FIELDS; ++i) {
    propptr[i] = prop[i];
  }
  *moneyptr = money;
  _last_offer = (is_src)? T_SRC : T_DEST;

//    assert(offer_is_valid(plr));

  if (!is_src) { // plr is dest
    _dest_replied_once = true;
  }

  notify_player(other, "Game: %s replied to your offer", plr->name());
  freelog(LOG_NORMAL, "%s replied to trade %d (%s to %s)", 
	  plr->name(), _id, srcplrname(), destplrname());
  
  game->send_trade(NULL, this, E_TRADE_REPLY, (is_src)?T_SRC:T_DEST);

  test_money_in_offer(srcplr);

  return 0;

}

void Trade::agree(Player *plr)
{
  assert(plr == srcplr || plr == destplr);

  int i;
  bool is_src = (plr == srcplr);
  Player *other = (is_src)?destplr:srcplr;

  // src can't agree before dest has replied once
  if (is_src && !_dest_replied_once) {
    notify_player(plr, "Game: wait until %s has made his offer", 
		  destplrname());
    freelog(LOG_NORMAL, "%s agreed in trade %d (%s to %s) but had no reply "
	    "of %s yet", 
	    srcplrname(), _id, srcplrname(), destplrname(), destplrname());
    return;
  }

  // which offer to take
  int *prop = (is_src)?_destprop:_srcprop;

  // test if offer is still valid, something could have been changed since 
  // the last offer
  if (!offer_is_valid(other)) {
    notify_player(plr, ("Game: The offer from %s is not valid.\n"
			"Game: Perhaps something has changed since his "
			"last offer. "),
		  other->name());
    notify_player(other, ("Game: %s wanted to agree with your offer but "
			  "your offer is not valid.\n"
			  "Game: Perhaps something has changed since your "
			  "last offer.\n"
			  "Game: Check money and properties involved in your "
			  "offer."),
		  plr->name());
    freelog(LOG_NORMAL, ("%s agreed in trade %d (%s to %s) but offer of %s was"
			 " invalid"),
	    plr->name(), _id, srcplrname(), destplrname(), other->name());
    return;
  }

  // we must check, if there are houses on the monopolies of traded streets
  // note: we could do this in offer or reply, but it is also possible,
  // that a player wants to wait selling the houses on a street till
  // the trade is completed

  for (i=0; i<BOARD_N_FIELDS; ++i) {

    if (prop[i] == T_NOTRADE) continue;

    // is field street ?
    Street *s = dynamic_cast<Street *>(game->board()->field(i));
    if (!s) continue;

    Monop *m = s->monop();
    
    if ( ( is_src  && (prop[i] & IS_SRC) ) ||
	 ( !is_src && !(prop[i] & IS_SRC) ) ) {	// plr gives prop[i]

      if (!m->no_houses()) {       // there are houses on monop
	notify_player(plr, ("Game: you have houses on any street in monopoly "
			    "%d, sell them before agreeing"),
		      m->id());
	freelog(LOG_NORMAL, "trade %d (%s to %s): %s has houses on some "
		"streets", _id, srcplrname(), destplrname(), plr->name());
	return;
      }

    } else {	// other gives prop[i]

      if (!m->no_houses()) {       // there are houses on monop
	notify_player(plr, ("Game: %s has houses on any street in monopoly %d"
			    ", wait for him to sell them"),
		      other->name(), m->id());
	notify_player(other, ("Game: %s agreed to your offer, but you have "
			      "houses on any street in monopoly %d.\n"
			      "Game: sell them before trading"),
		      plr->name(), m->id());
	freelog(LOG_NORMAL, "trade %d (%s to %s): %s has houses on some "
		"streets", _id, srcplrname(), destplrname(), other->name());
	return;
      }

    }
  }

  if (plr == srcplr) {

    _finishedflag = T_SRC_AGREED;
    _last_offer = T_DEST;	// the offer of destplr is taken

    freelog(LOG_NORMAL, "%s agreed in trade %d (%s to %s)", 
	    srcplrname(), _id, srcplrname(), destplrname());

  } else if (plr == destplr) {

    _finishedflag = T_DEST_AGREED;
    _last_offer = T_SRC;	// the offer of destplr is taken

    freelog(LOG_NORMAL, "%s agreed in trade %d (%s to %s)", 
	    destplrname(), _id, srcplrname(), destplrname());

  }

  
  finish_trade();

}

void Trade::reject(Player *plr)
{
  assert(plr == srcplr || plr == destplr);

  if (plr == srcplr) {

    _finishedflag = T_SRC_REJECTED;
    _last_offer = T_DEST;

//      notify_player(destplr, "Game: %s rejected your offer." , srcplrname());
    freelog(LOG_NORMAL, "%s rejected trade %d (%s to %s)", 
	    srcplrname(), _id, srcplrname(), destplrname());

  } else if (plr == destplr) {

    _finishedflag = T_DEST_REJECTED;
    _last_offer = T_SRC;

//      notify_player(srcplr, "Game: %s rejected your offer." , destplrname());
    freelog(LOG_NORMAL, "%s rejected trade %d (%s to %s)", 
	    destplrname(), _id, srcplrname(), destplrname());

  }

  game->send_trade(NULL, this, E_TRADE_REJECT, (plr==srcplr)?T_DEST:T_SRC);

}

bool Trade::offer_is_valid(Player *fromplr)
{
  int i, newstatus;

  // which offer to check
  Player *other = (fromplr == srcplr) ? destplr : srcplr;
  const int *prop = (fromplr == srcplr) ? _srcprop : _destprop;
  const int money = (fromplr == srcplr) ? _srcmoney : _destmoney;

  freelog(LOG_VERBOSE, "check validity of %s offer in trade %d (%s to %s) ", 
	  (fromplr == srcplr)?"src":"dest", _id, srcplrname(), destplrname());

  // check money
  if (money >= 0) { // fromplr has to pay
    if (!fromplr->has_enough_money(money)) {
      freelog(LOG_VERBOSE, "check failed: plr has not enough money");
      return false;
    }
  } else { // other has to pay
    if (!other->has_enough_money(-money)) {
      freelog(LOG_VERBOSE, "check failed: other has not enough money"); 
      return false;
    }
  }

  // check properties
  for (i=0; i<BOARD_N_FIELDS; ++i) {
    newstatus = game->board()->field(i)->status(srcplr);
    if (prop[i] != T_NOTRADE && prop[i] != newstatus) {
      // field i is involved in trade and status of field i has changed
      freelog(LOG_VERBOSE, ("check failed: status of field %d has changed "
			    "since offer"), i); 
      return false;
    }
  }
  
  // here all tests were successful
  freelog(LOG_VERBOSE, "check successful: offer is valid");
  return true;

}

void Trade::test_money_in_offer(Player *plr)
{
  assert(plr == srcplr || plr == destplr);

  // is plr src or dest ?
  bool is_src = (plr == srcplr)? true : false;
  Player *other = (is_src) ? destplr: srcplr;
  int *money = (is_src) ? &_srcmoney : &_destmoney;

  if (*money >= 0 && !plr->has_enough_money(*money)) { 
    // plr has to pay the money and currently can not
    // --> warning
    notify_player(plr, 
		  "Warning: You haven't got enough money for this trade.");
    notify_player(other, 
		  "Warning: %s hasn't got enough money for this trade.",
		  plr->name());
  } else if (*money < 0 && !other->has_enough_money(-*money)) {
    // other has to pay money and currently can not
    // --> warning
    notify_player(plr,
		  "Warning: %s hasn't got enough money for this trade.",
		  other->name());
    notify_player(other,
		  "Warning: You haven't got enough money for this trade.");
  }
}

bool const Trade::src_gives(int *prop, int i) const
{ 
  return ( (prop[i] != T_NOTRADE) && (prop[i] & IS_SRC) ) ? true : false;
}

bool const Trade::dest_gives(int *prop, int i) const
{
  return ( (prop[i] != T_NOTRADE) && !(prop[i] & IS_SRC) ) ? true : false;
}


int Trade::finish_trade()
{
  int i;

  bool srcagreed = (_finishedflag == T_SRC_AGREED);

  // which offer is taken, srcprop or destprop ?
  int *prop = (srcagreed)?_destprop:_srcprop;
  int money = (srcagreed)?_destmoney:_srcmoney;

  for (i=0; i<BOARD_N_FIELDS; ++i) {
    Property *p = dynamic_cast<Property *>(game->board()->field(i));
    if (src_gives(prop, i)) {
      p->change_owner(destplr);
    } else if (dest_gives(prop, i)) {
      p->change_owner(srcplr);
    }
  }

  srcplr->update_holdings();
  destplr->update_holdings();

  if (money > 0) {
    srcplr->pay(destplr, money, E_NOEVENT);
  } else if (money < 0) {
    destplr->pay(srcplr, -money, E_NOEVENT);
  }

  game->send_trade(NULL, this, E_TRADE_FINISHED, (srcagreed)?T_DEST:T_SRC);

  game->send_player_info(NULL, srcplr, E_TRADE_FINISHED);
  game->send_player_info(NULL, destplr, E_TRADE_FINISHED);

  freelog(LOG_NORMAL, "trade %d finished (%s to %s)", 
	  _id, srcplrname(), destplrname());
  return 0;
}
  
void Trade::prepare_packet_for(Player *, struct packet_trade *pack, int event,
			       int take_offer)
{
  int i;
  assert(take_offer == T_SRC || take_offer == T_DEST);
  int *prop = (take_offer == T_SRC) ? _srcprop:_destprop; 

  pack->tradeid = _id;

  pack->state = _finishedflag;
  pack->event = event;

  pack->srcid = srcplr->id();
  pack->destid = destplr->id();

  pack->offer_from = take_offer;

  pack->money = (take_offer == T_SRC) ? _srcmoney : _destmoney;

  for (i=0; i<BOARD_N_FIELDS; ++i) {
    // write T_NOTRADE, T_SRC or T_DEST instead of the statusflags stored
    // in the prop arrays
    if (prop[i] == T_NOTRADE) pack->prop[i] = T_NOTRADE;
    else if (prop[i] & IS_SRC) pack->prop[i] = T_SRC;
    else // IS_DEST
      pack->prop[i] = T_DEST;
  }
}



