// -*- C++ -*-
/***************************************************************************
                          stdinhand.cpp  -  description
                             -------------------
    begin                : Fre Jun 22 17:19:03 CEST 2001
    copyright            : (C) 2001 by Daniel Hermann
    email                : Daniel.Hermann@stud.uni-karlsruhe.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

//
// System
//

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <assert.h>
#include <stdarg.h>

//
// Common + Handler
//
#include "global.h"
#include "events.h"
#include "log.h"
#include "mem.h"
//#include "shared.h"

#include "main.h"
#include "sernet.h"

//#include "commands.h"
#include "settings.h"

#include "console.h"
#include "gamelog.h"
#include "plrhand.h"
#include "stdinhand.h"

//
// Game
//

#include "game.h"
#include "player.h"



#define MAX_CMD_LEN MAX_PACKET_SIZE
  /* to be used more widely - rp */

extern int gamelog_level;

extern Game *game;

enum cmdlevel_id default_access_level = ALLOW_INFO;

void cut_player_connection(Player *caller, char *playername);
void quit_game(Player *caller);
void show_help(Player *caller);
void show_players(Player *caller);

static char horiz_line[] =
"-------------------------------------------------------------------";

typedef enum {
    PNameOk,
    PNameEmpty,
    PNameTooLong
} PlayerNameStatus;

/**************************************************************************
...
**************************************************************************/
PlayerNameStatus test_player_name(char* name)
{
  int len = strlen(name);

  if (!len) {
      return PNameEmpty;
  } else if (len > 9) {
      return PNameTooLong;
  }

  return PNameOk;
}

/**************************************************************************
  Commands - can be recognised by unique prefix
**************************************************************************/
struct command {
  char *name;              /* name - will be matched by unique prefix   */
  enum cmdlevel_id level;  /* access level required to use the command  */
};

/* Order here is important: for ambiguous abbreviations the first
   match is used.  Arrange order to:
   - allow commands 's', 'h', 'l', 'q', 'c' to work.
   - reduce harm for ambiguous cases, where "harm" includes inconvenience,
     eg accidently removing a player in a running game.
*/
enum command_id {
  /* one-letter commands: */
  CMD_START = 0,
  CMD_HELP,
  CMD_LIST,
  CMD_SHOW,
  CMD_QUIT,

  /* mostly non-harmful: */
  CMD_SET,
  CMD_AITOGGLE,
  CMD_CMDLEVEL,

  /* undocumented */
  CMD_LOG,
  CMD_CRASH,

  /* pseudo-commands: */
  CMD_NUM,		/* the number of commands - for iterations */
  CMD_UNRECOGNIZED,	/* used as a possible iteration result */
  CMD_AMBIGUOUS		/* used as a possible iteration result */
};

static struct command commands[] = {
  {"start",	ALLOW_CTRL}, 
  {"help",	ALLOW_INFO},
  {"list",	ALLOW_INFO},
  {"show",	ALLOW_INFO},
  {"quit",	ALLOW_HACK},
  {"set",	ALLOW_CTRL},
  {"aitoggle",	ALLOW_CTRL},
  {"cmdlevel",	ALLOW_HACK},  /* confusing to leave this at ALLOW_CTRL */
  {"log",	ALLOW_HACK},
  {"crash",	ALLOW_HACK}
};

/**************************************************************************
  Convert a named command into an id.
  If accept_ambiguity is true, return the first command in the
  enum list which matches, else return CMD_AMBIGOUS on ambiguity.
  (This is a trick to allow ambiguity to be handled in a flexible way
  without importing notify_player() messages inside this routine - rp)
**************************************************************************/
static enum command_id command_named(char *token, int accept_ambiguity)
{
  int i;
  enum command_id found = CMD_UNRECOGNIZED;
  unsigned int len = strlen(token);

  for (i = 0; i < CMD_NUM; ++i) {
    if (strncmp(commands[i].name, token, len) == 0) {
      if (accept_ambiguity) {
	return (enum command_id) i;
      } else if (found != CMD_UNRECOGNIZED) {
        return CMD_AMBIGUOUS;
      } else if (strlen(commands[i].name) == len) {
	return (enum command_id) i;
      } else {
        found = (enum command_id) i;
      }
    }
  }
  return found;
}


/**************************************************************************
  Return current command access level for this player, or default if the
  player is not connected.
**************************************************************************/
static enum cmdlevel_id access_level(Player *pplayer)
{
  if (pplayer->is_connected()) {
    return pplayer->conn()->access_level;
  } else {
    return default_access_level;
  }
}

/**************************************************************************
  Whether the player can use the specified command.
  pplayer == NULL means console.
**************************************************************************/
static int may_use(Player *pplayer, enum command_id cmd)
{
  if (pplayer == NULL) {
    return 1;  /* on the console, everything is allowed */
  }
  return (access_level(pplayer) >= commands[cmd].level);
}

/**************************************************************************
  Whether the player cannot use any commands at all.
  pplayer == NULL means console.
**************************************************************************/
static int may_use_nothing(Player *pplayer)
{
  if (pplayer == NULL) {
    return 0;  /* on the console, everything is allowed */
  }
  return (access_level(pplayer) == ALLOW_NONE);
}

/********************************************************************
Returns whether the specified server setting (option) can currently
be changed.  Does not indicate whether it can be changed by clients.
*********************************************************************/
bool sset_is_changeable(int idx)
{
  struct settings_s *op = &settings[idx];

  switch(op->sclass) {
  case SSET_PLAYERS:
  case SSET_GAME_INIT:
  case SSET_RULES:
    /*
     * Only change start params and most rules if game hasn't started yet
     */
    return (server_state < RUN_GAME_STATE);
  case SSET_META:
    /* These can always be changed: */
    return true;
  default:
    freelog(LOG_FATAL, "Unexpected case %d in %s line %d",
            op->sclass, __FILE__, __LINE__);
    return false;
  }
}

/**************************************************************************
  Whether the caller can set the specified option (assuming that
  the state of the game would allow changing the option at all).
  caller == NULL means console.
**************************************************************************/
static bool may_set_option(Player *caller, int option_idx)
{
  if (!caller) {
    return true;  /* on the console, everything is allowed */
  } else {
    return (access_level(caller) == ALLOW_HACK);
  }
}

/**************************************************************************
  Whether the caller can set the specified option, taking into account
  access, and the game state.  caller == NULL means console.
**************************************************************************/
static bool may_set_option_now(Player *caller, int option_idx)
{
  return (may_set_option(caller, option_idx)
	  && sset_is_changeable(option_idx));
}

/**************************************************************************
  Whether the caller can SEE the specified option.
  caller == NULL means console, which can see all.
  client players can see "to client" options, or if player
  has command access level to change option.
**************************************************************************/
static bool may_view_option(Player *caller, int option_idx)
{
  if (!caller) {
    return true;  /* on the console, everything is allowed */
  } else {
    return may_set_option(caller, option_idx);
  }
}

/**************************************************************************
  feedback related to server commands
  caller == NULL means console.
  No longer duplicate all output to console.
  'console_id' should be one of the C_ identifiers in console.h
**************************************************************************/
static void cmd_reply(enum command_id cmd, Player *caller,
		      int console_id, char *format, ...)
{
  char line[MAX_CMD_LEN];
  va_list ap;
  const char *cmdname = cmd < CMD_NUM ? commands[cmd].name :
                  cmd == CMD_AMBIGUOUS ? "(ambiguous)" :
                  cmd == CMD_UNRECOGNIZED ? "(unknown)" :
			"(?!?)";  /* this case is a bug! */

  va_start(ap,format);
  /* (void)vsnprintf(line,MAX_CMD_LEN-1,format,ap); */
    /*
     * no snprintf() in ANSI C and I don't know how to do it otherwise;
     * then again, there are plenty of other places in which this overflow
     * exists so I won't solve it here until we have myvsnprintf() - rp
     */
  vsprintf(line,format,ap);
  va_end(ap);

  if (caller) {
    notify_player(caller, "/%s: %s", const_cast<char *>(cmdname), line);
    /* cc: to the console - testing has proved it's too verbose - rp
    con_write(console_id, "%s/%s: %s", caller->name, cmdname, line);
    */
  } else {
    con_write(console_id, "%s", line);
  }
}

/******************************************************************
  Which characters are allowed within option names: (for 'set')
******************************************************************/
static bool is_ok_opt_name_char(char c)
{
  return isalnum(c);
}

/******************************************************************
  Which characters are allowed within option values: (for 'set')
******************************************************************/
static bool is_ok_opt_value_char(char c)
{
  return (c == '-') || (c == '*') || (c == '+') || (c == '=') || isalnum(c);
}

/******************************************************************
  Which characters are allowed between option names and values: (for 'set')
******************************************************************/
static bool is_ok_opt_name_value_sep_char(char c)
{
  return (c == '=') || isspace(c);
}

/**************************************************************************
Find option index by name. Return index (>=0) on success, -1 if no
suitable options were found, -2 if several matches were found.
**************************************************************************/
#define OPT_UNRECOGNIZED -1
#define OPT_AMBIGUOUS -2
static int lookup_option(const char *name)
{
  int i;
  int found = OPT_UNRECOGNIZED;
  unsigned int len = strlen(name);

  for (i = 0; i < SETTINGS_NUM; ++i) {
    if (strncmp(settings[i].name, name, len) == 0) {
      if (found != OPT_UNRECOGNIZED) {
        return OPT_AMBIGUOUS;  // AMBIGUOUS
      } else if (strlen(settings[i].name) == len) {
	return i;
      } else {
        found = i;
      }
    }
  }
  return found;
}

/******************************************************************
Print a summary of the settings and their values.
Note that most values are at most 4 digits, except seeds,
which we let overflow their columns, plus a sign character.
Only show options which the caller can SEE.
******************************************************************/
void show_command(Player *caller, char *str)
{
  char buf[MAX_CMD_LEN];
  char command[MAX_CMD_LEN], *cptr_s, *cptr_d;
  int cmd,i,len1;
  size_t clen = 0;

  for (cptr_s = str; *cptr_s != '\0' && !isalnum(*cptr_s); cptr_s++) {
    /* nothing */
  }
  for (cptr_d = command; *cptr_s != '\0' && isalnum(*cptr_s); cptr_s++, cptr_d++)
    *cptr_d=*cptr_s;
  *cptr_d='\0';

  if (*command != '\0') {
    cmd=lookup_option(command);
    if (cmd >= 0) {

      if (!may_view_option(caller, cmd)) {
        cmd_reply(CMD_SHOW, caller, C_FAIL,
		  "Sorry, you do not have access to view option '%s'.",
		  command);
        return;
      }
    }
    if (cmd == OPT_UNRECOGNIZED) {
      cmd_reply(CMD_SHOW, caller, C_FAIL, "Unknown option '%s'.", command);
      return;
    }
    if (cmd == OPT_AMBIGUOUS) {
      /* allow ambiguous: show all matching */
      clen = strlen(command);
    }
  } else {
   cmd = OPT_UNRECOGNIZED;  /* to indicate that no comannd was specified */
  }

#define cmd_reply_show(string)  cmd_reply(CMD_SHOW, caller, C_COMMENT, string)

#define OPTION_NAME_SPACE 13
  /* under SSET_MAX_LEN, so it fits into 80 cols more easily - rp */

  cmd_reply_show(horiz_line);
  cmd_reply_show("+ means you may change the option");
  cmd_reply_show("= means the option is on its default value");
  cmd_reply_show(horiz_line);
  len1 = snprintf(buf, sizeof(buf),
	"%-*s value   (min,max)      ", OPTION_NAME_SPACE, "Option");
  if (len1 == -1)
    len1 = sizeof(buf) -1;
  sz_strlcat(buf, "description");
  cmd_reply_show(buf);
  cmd_reply_show(horiz_line);

  buf[0] = '\0';

  for (i = 0; settings[i].name; i++) {
    if (may_view_option(caller, i)
	&& (cmd == OPT_UNRECOGNIZED || cmd == i
	    || (cmd == OPT_AMBIGUOUS
		&& mystrncasecmp(settings[i].name, command, clen) == 0))) {
      /* in the cmd==i case, this loop is inefficient. never mind - rp */
      struct settings_s *op = &settings[i];
      int len = 0;

      if (true) {
        switch (op->type) {
        case SSET_BOOL:
	  len = snprintf(buf, sizeof(buf),
			    "%-*s %c%c%-5d (0,1)", OPTION_NAME_SPACE, op->name,
			    may_set_option_now(caller, i) ? '+' : ' ',
			    ((*op->bool_value == op->bool_default_value) ?
			     '=' : ' '), (*op->bool_value) ? 1 : 0);
	  break;

        case SSET_INT:
	  len = snprintf(buf, sizeof(buf),
			    "%-*s %c%c%-5d (%d,%d)", OPTION_NAME_SPACE,
			    op->name, may_set_option_now(caller,
						         i) ? '+' : ' ',
			    ((*op->int_value == op->int_default_value) ?
			     '=' : ' '),
			    *op->int_value, op->int_min_value,
			    op->int_max_value);
	  break;

        case SSET_STRING:
	  len = snprintf(buf, sizeof(buf),
			    "%-*s %c%c\"%s\"", OPTION_NAME_SPACE, op->name,
			    may_set_option_now(caller, i) ? '+' : ' ',
			    ((strcmp(op->string_value,
				     op->string_default_value) == 0) ?
			     '=' : ' '), op->string_value);
	  break;
        }
        if (len == -1) {
          len = sizeof(buf) - 1;
        }
        /* Line up the descriptions: */
        if (len < len1) {
          cat_snprintf(buf, sizeof(buf), "%*s", (len1-len), " ");
        } else {
          sz_strlcat(buf, " ");
        }
        // sz_strlcat(buf, _(op->short_help));
        cmd_reply_show(buf);
      }
    }
  }
  cmd_reply_show(horiz_line);
#undef cmd_reply_show
#undef OPTION_NAME_SPACE
}

/******************************************************************
  ...
******************************************************************/
void set_command(Player *caller, char *arg)
{
  char opt_str[MAX_CMD_LEN+1], val_str[MAX_CMD_LEN+1], *cptr_s, *cptr_d;
  int val, opt_idx;
  struct settings_s *op;
  bool do_update;
  char buffer[500];

  for (cptr_s = arg; *cptr_s != '\0' && !is_ok_opt_name_char(*cptr_s);
       cptr_s++) {
    /* nothing */
  }

  for(cptr_d=opt_str;
      *cptr_s != '\0' && is_ok_opt_name_char(*cptr_s);
      cptr_s++, cptr_d++) {
    *cptr_d=*cptr_s;
  }
  *cptr_d='\0';
  
  for (; *cptr_s != '\0' && is_ok_opt_name_value_sep_char(*cptr_s); cptr_s++) {
    /* nothing */
  }

  for (cptr_d = val_str; *cptr_s != '\0' && is_ok_opt_value_char(*cptr_s); cptr_s++, cptr_d++)
    *cptr_d=*cptr_s;
  *cptr_d='\0';

  opt_idx = lookup_option(opt_str);
  if (opt_idx==OPT_UNRECOGNIZED) {
    cmd_reply(CMD_SET, caller, C_SYNTAX,
	      "Undefined option name.  Usage: set <option> <value>.");
    return;
  }
  else if (opt_idx==OPT_AMBIGUOUS) {
    cmd_reply(CMD_SET, caller, C_SYNTAX,
	      "Ambiguous option name.");
    return;
  }
  if (!may_set_option(caller,opt_idx)) {
     cmd_reply(CMD_SET, caller, C_FAIL,
	       "You are not allowed to set this option.");
    return;
  }
  if (!sset_is_changeable(opt_idx)) {
    cmd_reply(CMD_SET, caller, C_BOUNCE,
	      "This setting can't be modified after the game has started.");
    return;
  }

  op = &settings[opt_idx];

  do_update = false;
  buffer[0] = '\0';

  switch (op->type) {
  case SSET_BOOL:
    if (sscanf(val_str, "%d", &val) != 1) {
      cmd_reply(CMD_SET, caller, C_SYNTAX, "Value must be an integer.");
      return;
    } else if (val != 0 && val != 1) {
      cmd_reply(CMD_SET, caller, C_SYNTAX,
		"Value out of range (minimum: 0, maximum: 1).");
      return;
    } else {
      const char *reject_message = NULL;
      bool b_val = (val != 0);

      if (settings[opt_idx].bool_validate
	  && !settings[opt_idx].bool_validate(b_val, &reject_message)) {
        // TODO: fix the typecast of reject_message, otherwise this will segfault!
	cmd_reply(CMD_SET, caller, C_SYNTAX, (char *) reject_message);
        return;
      } else {
	*(op->bool_value) = b_val;
	snprintf(buffer, sizeof(buffer),
		    "Option: %s has been set to %d.", op->name,
		    *(op->bool_value) ? 1 : 0);
	do_update = true;
      }
    }
    break;

  case SSET_INT:
    if (sscanf(val_str, "%d", &val) != 1) {
      cmd_reply(CMD_SET, caller, C_SYNTAX, "Value must be an integer.");
      return;
    } else if (val < op->int_min_value || val > op->int_max_value) {
      cmd_reply(CMD_SET, caller, C_SYNTAX,
		"Value out of range (minimum: %d, maximum: %d).",
		op->int_min_value, op->int_max_value);
      return;
    } else {
      const char *reject_message = NULL;

      if (settings[opt_idx].int_validate
	  && !settings[opt_idx].int_validate(val, &reject_message)) {
	cmd_reply(CMD_SET, caller, C_SYNTAX, (char *) reject_message);
        return;
      } else {
	*(op->int_value) = val;
	snprintf(buffer, sizeof(buffer),
		    "Option: %s has been set to %d.", op->name,
		    *(op->int_value));
	do_update = true;
      }
    }
    break;

  case SSET_STRING:
    if (strlen(val_str) >= op->string_value_size) {
      cmd_reply(CMD_SET, caller, C_SYNTAX,
		"String value too long.  Usage: set <option> <value>.");
      return;
    } else {
      const char *reject_message = NULL;

      if (settings[opt_idx].string_validate
	  && !settings[opt_idx].string_validate(val_str, &reject_message)) {
	cmd_reply(CMD_SET, caller, C_SYNTAX, (char *) reject_message);
        return;
      } else {
	strcpy(op->string_value, val_str);
	snprintf(buffer, sizeof(buffer),
		    "Option: %s has been set to \"%s\".", op->name,
		    op->string_value);
	do_update = true;
      }
    }
    break;
  }

  if (strlen(buffer) > 0) {
    notify_player(0,"%s", buffer);
  }

  if (do_update) {
    /* 
     * send any modified game parameters to the clients -- if sent
     * before RUN_GAME_STATE, triggers a popdown_races_dialog() call
     * in client/packhand.c#handle_game_info() 
     */
    if (server_state == RUN_GAME_STATE) {
      game->send_game_info(TOALL);
    }
  }
}

/**************************************************************************
...
**************************************************************************/
static void toggle_ai_player_direct(Player *caller, Player *pplayer)
{
  assert(pplayer);

  // toggle the AI state of the player
  pplayer->set_ai_controlled(!pplayer->is_ai_controlled());

  if (pplayer->is_ai_controlled()) {
    notify_player(0,"Game: player %s is now under AI control.", pplayer->name());
  } else {
    cmd_reply(CMD_AITOGGLE, caller, C_OK,
	      "%s is now under human control.", pplayer->name());
    notify_player(0,"Game: player %s is now under human control.", pplayer->name());
  }

  const char *new_status = pplayer->is_ai_controlled() ? "AI controlled" 
                                                 : "under human control";
  cmd_reply(CMD_AITOGGLE, caller, C_OK,
	    "%s is now %s.", pplayer->name(), (char *) new_status);
  freelog(LOG_NORMAL, "%s is now %s", pplayer->name(), new_status);
  notify_player(TOALL, "Game: %s is now %s.", pplayer->name(), new_status);
  game->send_player_info(TOALL, pplayer, E_TOGGLE_AI);
}

/**************************************************************************
...
**************************************************************************/
void toggle_ai_player(Player *caller, char *name)
{
  Player *pplayer;
  if (test_player_name(name) == PNameOk &&
      (pplayer=game->find_player_by_name(name))) {
    toggle_ai_player_direct(caller, pplayer);
  } else {
    cmd_reply(CMD_AITOGGLE, caller, C_FAIL,
	      "cannot toggle ai for unknown/invalid player "
	      "name '%s'", name);
  }
}

/**************************************************************************
 set pplayer's cmdlevel to level if caller is allowed to do so 
**************************************************************************/
static int set_cmdlevel(Player *caller, Player *pplayer,
			enum cmdlevel_id level)
{
  assert(pplayer && pplayer->conn());
  /* only ever call me for specific, connected players */

  if (caller && access_level(pplayer) > access_level(caller)) {
    /* Can this happen?  Caller must already have ALLOW_HACK
       to even use the cmdlevel command.  Hmm, someone with
       ALLOW_HACK can take away ALLOW_HACK from others... --dwp
    */
    cmd_reply(CMD_CMDLEVEL, caller, C_FAIL,
	      "cannot decrease command level '%s' for player '%s',"
	      " you only have '%s'",
	      cmdlevel_name(access_level(pplayer)),
	      pplayer->name(),
	      cmdlevel_name(access_level(caller)));
    return 0;
  } else {
    pplayer->conn()->access_level = level;
    notify_player(pplayer, "Game: you now have access level '%s'",
		  cmdlevel_name(level));
    return 1;
  }
}

/**************************************************************************
 Change command level for individual player, or all, or new.
**************************************************************************/
void cmdlevel_command(Player *caller, char *str)
{
  char arg_level[MAX_CMD_LEN+1]; /* info, ctrl etc */
  char arg_name[MAX_CMD_LEN+1];	 /* a player name, or "new" */
  char *cptr_s, *cptr_d;	 /* used for string ops */

  enum cmdlevel_id level;
  Player *pplayer;

  /* find the start of the level: */
  for(cptr_s=str; *cptr_s && !isalnum(*cptr_s); cptr_s++);

  /* copy the level into arg_level[] */
  for(cptr_d=arg_level; *cptr_s && isalnum(*cptr_s); cptr_s++, cptr_d++) {
    *cptr_d=*cptr_s;
  }
  *cptr_d='\0';
  
  if (!arg_level[0]) {
    /* no level name supplied; list the levels */
    int i;

    cmd_reply(CMD_CMDLEVEL, caller, C_COMMENT, "Command levels in effect:");

    for (i = 0; i < game->nplayers(); ++i) {
      Player *pplayer = game->player(i);
      if (pplayer->conn()) {
	cmd_reply(CMD_CMDLEVEL, caller, C_COMMENT,
		  "cmdlevel %s %s",
		  cmdlevel_name(access_level(pplayer)), pplayer->name());
      } else {
	cmd_reply(CMD_CMDLEVEL, caller, C_COMMENT,
		  "cmdlevel %s %s (not connected)",
		  cmdlevel_name(ALLOW_NONE), pplayer->name());
      }
    }
    cmd_reply(CMD_CMDLEVEL, caller, C_COMMENT,
	      "Default command level for new connections: %s",
	      cmdlevel_name(default_access_level));
    return;
  }

  /* a level name was supplied; set the level */

  if ((level = cmdlevel_named(arg_level)) == ALLOW_UNRECOGNIZED) {
    cmd_reply(CMD_CMDLEVEL, caller, C_SYNTAX,
	      "error: command level must be one of 'none', 'info', 'ctrl', "
	      "or 'hack'");
    return;
  } else if (caller && level > access_level(caller)) {
    cmd_reply(CMD_CMDLEVEL, caller, C_FAIL,
	      "cannot increase command level to '%s', you only have '%s' "
	      "yourself", arg_level, cmdlevel_name(access_level(caller)));
    return;
  }

  /* find the start of the name: */
  for(; *cptr_s && !isalnum(*cptr_s); cptr_s++);

  /* copy the name into arg_name[] */
  for(cptr_d=arg_name; *cptr_s && (*cptr_s == '-' || isalnum(*cptr_s));
      cptr_s++ , cptr_d++) {
    *cptr_d=*cptr_s;
  }
  *cptr_d='\0';
 
  if (!arg_name[0]) {
    /* no playername supplied: set for all connected players,
     * and set the default
     */
    int i;
    for (i = 0; i < game->nplayers(); ++i) {
      Player *pplayer = game->player(i);
      if (pplayer->conn()) {
	if (set_cmdlevel(caller, pplayer, level)) {
	  cmd_reply(CMD_CMDLEVEL, caller, C_OK,
		    "command access level set to '%s' for player %s",
		    cmdlevel_name(level),
		    pplayer->name());
	} else {
	  cmd_reply(CMD_CMDLEVEL, caller, C_OK,
		    "command access level could not be set to '%s' for "
		    "player %s", cmdlevel_name(level), pplayer->name());
	}
      }
    }
    default_access_level = level;
    cmd_reply(CMD_CMDLEVEL, caller, C_OK,
	      "default command access level set to '%s'",
	      cmdlevel_name(level));
    notify_player(0,"Game: all players now have access level '%s'",
		  cmdlevel_name(level));
  }
  else if (strcmp(arg_name,"new") == 0) {
    default_access_level = level;
    cmd_reply(CMD_CMDLEVEL, caller, C_OK,
	      "default command access level set to '%s'",
	      cmdlevel_name(level));
    notify_player(0,"Game: new connections will have access level '%s'",
		  cmdlevel_name(level));
  }
  else if (test_player_name(arg_name) == PNameOk &&
	   (pplayer=game->find_player_by_name(arg_name))) {
    if (!pplayer->conn()) {
      cmd_reply(CMD_CMDLEVEL, caller, C_FAIL,
		"cannot change command access for unconnected player '%s'",
		arg_name);
      return;
    }
    if (set_cmdlevel(caller,pplayer,level)) {
      cmd_reply(CMD_CMDLEVEL, caller, C_OK,
		"command access level set to '%s' for player %s",
		cmdlevel_name(level), pplayer->name());
    } else {
      cmd_reply(CMD_CMDLEVEL, caller, C_OK,
		"command access level could not be set to '%s' for player %s",
		cmdlevel_name(level), pplayer->name());
    }
  } else {
    cmd_reply(CMD_CMDLEVEL, caller, C_FAIL,
	      "cannot change command access for unknown/invalid player "
	      "name '%s'", arg_name);
  }
}


void crash_and_burn(Player *caller)
{
  cmd_reply(CMD_CRASH, caller, C_GENFAIL, "Crashing and burning.");
  /* Who is General Failure and why is he crashing and
     burning my computer? :) -- Per */
   assert(0);
}

/**************************************************************************
...
**************************************************************************/
void handle_stdin_input(Player *caller, char *str)
{
  char command[MAX_CMD_LEN+1], arg[MAX_CMD_LEN+1], *cptr_s, *cptr_d;
  int i;
  enum command_id cmd;

  /* notify to the server console */
  if (caller) {
    con_write(C_COMMENT, "%s: '%s'", caller->name(), str);
  }

  /* if the caller may not use any commands at all, don't waste any time */
  if (may_use_nothing(caller)) {
    cmd_reply(CMD_HELP, caller, C_FAIL,
	"Sorry, you are not allowed to use server commands.");
     return;
  }

  /* Is it a comment or a blank line? */
  /* line is comment if the first non-whitespace character is '#': */
  for(cptr_s=str; *cptr_s && isspace(*cptr_s); cptr_s++);
  if(*cptr_s == 0 || *cptr_s == '#') {
    return;
  }

  /* commands may be prefixed with SERVER_COMMAND_PREFIX,
     even when given on the server command line - rp */

  if (*cptr_s == SERVER_COMMAND_PREFIX) cptr_s++;

  for(cptr_s=str; *cptr_s && !isalnum(*cptr_s); cptr_s++);

  for(cptr_d=command; *cptr_s && isalnum(*cptr_s) &&
      cptr_d < command+sizeof(command)-1; cptr_s++, cptr_d++)
    *cptr_d=*cptr_s;
  *cptr_d='\0';

  cmd = command_named(command,0);
  if (cmd == CMD_AMBIGUOUS) {
    cmd = command_named(command,1);
    cmd_reply(cmd, caller, C_SYNTAX,
	"Warning: '%s' interpreted as '%s', but it is ambiguous. Try '%shelp'.",
	command, commands[cmd].name, caller?"/":"");
  } else if (cmd == CMD_UNRECOGNIZED) {
    cmd_reply(cmd, caller, C_SYNTAX,
	"Unknown command.  Try '%shelp'.", caller?"/":"");
    return;
  }
  if (!may_use(caller, cmd)) {
    assert(caller);
    cmd_reply(cmd, caller, C_FAIL, "You are not allowed to use this command.");
    return;
  }

  for(; *cptr_s && isspace(*cptr_s); cptr_s++);
  strncpy(arg, cptr_s, MAX_CMD_LEN);
  arg[MAX_CMD_LEN]='\0';

  i=strlen(arg)-1;
  while(i>0 && isspace(arg[i]))
    arg[i--]='\0';

  switch(cmd) {
  case CMD_HELP:
    show_help(caller);
    break;
  case CMD_LIST:
    show_players(caller);
    break;
  case CMD_SHOW:
    show_command(caller, arg);
    break;
  case CMD_CRASH:
    crash_and_burn(caller);
  case CMD_LOG:		/* undocumented */
    freelog(LOG_NORMAL, "%s", arg);
    break;
  case CMD_QUIT:
    quit_game(caller);
    break;
  case CMD_START:
    if(server_state==PRE_GAME_STATE) {
      int plrs=0;
      int i;

      for (i=0;i<game->nplayers();i++) {
        if (game->player(i)->is_connected()) plrs++ ;
      }

//       if (plrs<GAME_MIN_PLAYERS) {
//         cmd_reply(cmd,caller, C_FAIL,
// 		  "Not enough players, game will not start.");
//       } else {
	cmd_reply(cmd, caller, C_OK, "starting game.");
	freelog(LOG_NORMAL, "game started by %s", 
		(caller)?caller->name():"operator");
	set_server_state(SELECT_TOKEN_STATE); 
//       }
    }
    else {
      cmd_reply(cmd, caller, C_FAIL,
		"Cannot start the game: it is already running.");
    }
    break;
  case CMD_SET:
    set_command(caller, arg);
    break;
  case CMD_AITOGGLE:
    toggle_ai_player(caller, arg);
    break;
  case CMD_CMDLEVEL:
    cmdlevel_command(caller,arg);
    break;
  case CMD_NUM:
  case CMD_UNRECOGNIZED:
  case CMD_AMBIGUOUS:
  default:
    freelog(LOG_FATAL, "bug in server: impossible command recognized; bye!");
    assert(0);
  }
}

/*
 * IO-functions
 */

int get_str(char *prompt, char *return_str)
{
  char buf[BUF_SIZE+1];

  con_dump(C_READY, prompt);
  if (fgets(buf, BUF_SIZE, stdin) == NULL) {
    return -1;
  }
  *(buf+BUF_SIZE) = '\0';	/* in case there is no '\0' in buf left */
  *(buf+strlen(buf)-1) = '\0';	/* get rid of '\n' */
  strcpy(return_str, buf);
  return strlen(return_str);
}

int getyn(char *prompt)
{
  char buf[BUF_SIZE+1];

  for (;;) {
    if (get_str(prompt, buf) < 0) {
      con_write(C_FAIL, "read from stdin failed");
      return -1;
    } else if (!strcmp(buf, "y")) {
      return 1;
    } else if (!strcmp(buf, "n")) {
      return 0;
    } else {
      con_write(C_FAIL, "Please answer y or n!");
    }
  }
}

/**************************************************************************
...
**************************************************************************/
void quit_game(Player *caller)
{
  int i;
  struct packet_generic_message gen_packet;
  gen_packet.message[0]='\0';
  
  i = getyn("Do you really want to quit ? ");

  if (i == 1) {
    for(i=0; i<game->nplayers(); i++)
      send_packet_generic_message(game->player(i)->conn(), 
				  PACKET_SERVER_SHUTDOWN,
				  &gen_packet);
    close_connections_and_socket();
    
    exit(0);
  }
}

/**************************************************************************
...
**************************************************************************/
void show_help(Player *caller)
{
  assert(!may_use_nothing(caller));
    /* no commands means no help, either */

#define cmd_reply_help(cmd,string) \
  if (may_use(caller,cmd)) \
    cmd_reply(CMD_HELP, caller, C_COMMENT, string)

  cmd_reply_help(CMD_HELP,
	"Available commands: (P=player, L=level, O=option V=value)");
  cmd_reply_help(CMD_HELP, horiz_line);
  cmd_reply_help(CMD_HELP, "help            - this help text");
  cmd_reply_help(CMD_QUIT, "quit            - quit game and shutdown server");
  cmd_reply_help(CMD_LIST, "list            - list players");
  cmd_reply_help(CMD_LIST, "show            - show available server options");
  cmd_reply_help(CMD_LIST, "set O V         - set server option X to value V");
  if(server_state==PRE_GAME_STATE) {
    cmd_reply_help(CMD_START, "start           - start game");
  }
  cmd_reply_help(CMD_CMDLEVEL, "cmdlevel        - see current command levels");
  cmd_reply_help(CMD_CMDLEVEL, "cmdlevel L      "
		 "- sets command access level to L for all players");
  cmd_reply_help(CMD_CMDLEVEL, "cmdlevel L new  "
		 "- sets command access level to L for new connections");
  cmd_reply_help(CMD_CMDLEVEL,
	"cmdlevel L P    - sets command access level to L for player P");
  cmd_reply_help(CMD_LIST, "aitoggle P      - toggle AI status of player P");
  cmd_reply_help(CMD_HELP,
	"Abbreviations are allowed.");
#undef cmd_reply_help
}

/**************************************************************************
...
**************************************************************************/
void show_players(Player *caller)
{
  int i;
  
  cmd_reply(CMD_LIST, caller, C_COMMENT, "List of players:");
  cmd_reply(CMD_LIST, caller, C_COMMENT, horiz_line);

  if (game->nplayers() == 0)
    cmd_reply(CMD_LIST, caller, C_WARNING, "<no players>");
  else
  {
    for(i=0; i<game->nplayers(); i++) {
      if (game->player(i)->is_connected()) {
	cmd_reply(CMD_LIST, caller, C_COMMENT,
		  "%s is connected from %s",
		  game->player(i)->name(),
		  game->player(i)->addr()); 
      } else {
	cmd_reply(CMD_LIST, caller, C_COMMENT,
		  "%s is not connected",
		  game->player(i)->name());
      }
    }
  }
  cmd_reply(CMD_LIST, caller, C_COMMENT, horiz_line);
}





